import numpy as np
import pytest
from astropy.utils.exceptions import AstropyUserWarning
from specutils import Spectrum1D


def test_linking_after_spectral_smooth(cubeviz_helper, spectrum1d_cube):
    app = cubeviz_helper.app
    dc = app.data_collection
    data_label = 'test'
    cubeviz_helper.load_data(spectrum1d_cube, data_label=data_label)
    spec_viewer = cubeviz_helper.app.get_viewer('spectrum-viewer')

    assert len(dc) == 2

    gs = cubeviz_helper.plugins['Gaussian Smooth']._obj
    gs.mode_selected = 'Spectral'
    assert len(gs.dataset_items) == 2  # cube and extracted spectrum
    # selecting the cube as input will result in an output cube
    gs.dataset_selected = f'{data_label}[FLUX]'
    gs.stddev = 3.2
    gs.add_to_viewer_selected = 'None'
    assert gs.results_label == f'{data_label}[FLUX] spectral-smooth stddev-3.2'
    gs.vue_apply()
    # when not showing the results, the label will remain the same,
    # so there should be an overwrite warning
    assert gs.results_label_overwrite is True
    gs.add_to_viewer_selected = 'uncert-viewer'
    gs.vue_apply()
    # but since we're overwriting, add_to_viewer_selected is ignored (and hidden in the UI)
    # so will still be hidden
    assert len(gs.dataset_items) == 2
    # by removing the data entry, the overwrite will no longer apply
    app.remove_data_from_viewer('uncert-viewer', f'{data_label}[FLUX] spectral-smooth stddev-3.2')
    app.data_collection.remove(
        app.data_collection[f'{data_label}[FLUX] spectral-smooth stddev-3.2'])

    gs.add_to_viewer_selected = 'uncert-viewer'
    gs.vue_apply()

    # FOR HISTORICAL CONTEXT:
    # The data label used to be prepended to the results_label ONLY if there were multiple
    # smoothed spectra. As of PR#1973, POs requested the data label to always be present.
    # As a result, label will overwrite here
    assert len(gs.dataset_items) == 2
    assert gs.dataset_selected == f'{data_label}[FLUX]'
    assert gs.results_label == f'{data_label}[FLUX] spectral-smooth stddev-3.2'
    assert gs.results_label_overwrite is True

    assert len(dc) == 3
    assert dc[-1].label == f'{data_label}[FLUX] spectral-smooth stddev-3.2'
    assert len(dc.external_links) == 5

    # Link cube 3D x, y, z to plugin 3D x, y, z

    # Link 2:
    # Pixel Axis 0 [z] from cube.pixel_component_ids[0]
    # Pixel Axis 0 [z] from plugin.pixel_component_ids[0]
    assert dc.external_links[2].cids1[0] == dc[0].pixel_component_ids[0]
    assert dc.external_links[2].cids2[0] == dc[-1].pixel_component_ids[0]

    # Link 3:
    # Pixel Axis 1 [y] from cube.pixel_component_ids[1]
    # Pixel Axis 1 [y] from plugin.pixel_component_ids[1]
    assert dc.external_links[3].cids1[0] == dc[0].pixel_component_ids[1]
    assert dc.external_links[3].cids2[0] == dc[-1].pixel_component_ids[1]

    # Link 4:
    # Pixel Axis 2 [x] from cube.pixel_component_ids[2]
    # Pixel Axis 2 [x] from plugin.pixel_component_ids[2]
    assert dc.external_links[4].cids1[0] == dc[0].pixel_component_ids[2]
    assert dc.external_links[4].cids2[0] == dc[-1].pixel_component_ids[2]

    # Spectral smooth the input SPECTRUM
    gs.dataset_selected = 'Spectrum (sum)'
    gs.vue_apply()

    # Mouseover should automatically jump from one spectrum
    # to another, depending on which one is closer.

    label_mouseover = cubeviz_helper.app.session.application._tools['g-coords-info']
    label_mouseover._viewer_mouse_event(spec_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 4.6236e-7, 'y': 60}})
    assert label_mouseover.as_text() == ('Cursor 4.62360e-07, 6.00000e+01',
                                         'Wave 4.62360e-07 m (1 pix)',
                                         'Flux 9.20000e+01 Jy')
    assert label_mouseover.icon == 'b'

    label_mouseover._viewer_mouse_event(spec_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 4.6236e-7, 'y': 20}})
    assert label_mouseover.as_text() == ('Cursor 4.62360e-07, 2.00000e+01',
                                         'Wave 4.62360e-07 m (1 pix)',
                                         'Flux 1.47943e+01 Jy')
    assert label_mouseover.icon == 'd'

    # Check mouseover behavior when we hide everything.
    for lyr in spec_viewer.layers:
        lyr.visible = False

    label_mouseover._viewer_mouse_event(spec_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 4.6236e-7, 'y': 60}})
    assert label_mouseover.as_text() == ('Cursor 4.62360e-07, 6.00000e+01', '', '')
    assert label_mouseover.icon == 'mdi-cursor-default'


def test_spatial_convolution(cubeviz_helper, spectrum1d_cube):
    data_label = 'test'
    dc = cubeviz_helper.app.data_collection
    cubeviz_helper.load_data(spectrum1d_cube, data_label=data_label)

    gs = cubeviz_helper.plugins['Gaussian Smooth']._obj
    gs.mode_selected = 'Spatial'
    assert len(gs.dataset_items) == 1
    gs.dataset_selected = f'{data_label}[FLUX]'
    gs.stddev = 3
    assert gs.results_label == f'{data_label}[FLUX] spatial-smooth stddev-3.0'
    with pytest.warns(
            AstropyUserWarning,
            match='The following attributes were set on the data object, but will be ignored'):
        gs.vue_apply()

    assert len(dc) == 3
    assert dc[-1].label == f'{data_label}[FLUX] spatial-smooth stddev-3.0'
    assert dc[-1].shape == (4, 2, 2)  # specutils moved spectral axis to last
    assert (dc[f'{data_label}[FLUX] spatial-smooth stddev-3.0'].get_object(cls=Spectrum1D,
                                                                           statistic=None).shape
            == (4, 2, 2))


def test_specviz_smooth(specviz_helper, spectrum1d):
    data_label = 'test'
    dc = specviz_helper.app.data_collection
    specviz_helper.load_data(spectrum1d, data_label=data_label)
    spec_viewer = specviz_helper.app.get_viewer('spectrum-viewer')

    gs = specviz_helper.plugins['Gaussian Smooth']._obj
    gs.dataset_selected = data_label
    gs.mode_selected = 'Spectral'
    gs.stddev = 10
    gs.vue_apply()

    assert len(dc) == 2
    assert dc[1].label == f'{data_label} smooth stddev-10.0'

    # Mouseover should automatically jump from one spectrum
    # to another, depending on which one is closer.

    label_mouseover = specviz_helper.app.session.application._tools['g-coords-info']
    label_mouseover._viewer_mouse_event(spec_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 6400, 'y': 120}})
    assert label_mouseover.as_text() == ('Cursor 6.40000e+03, 1.20000e+02',
                                         'Wave 6.44444e+03 Angstrom (2 pix)',
                                         'Flux 1.35366e+01 Jy')
    assert label_mouseover.icon == 'a'

    label_mouseover._viewer_mouse_event(spec_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 6400, 'y': 5}})
    assert label_mouseover.as_text() == ('Cursor 6.40000e+03, 5.00000e+00',
                                         'Wave 6.44444e+03 Angstrom (2 pix)',
                                         'Flux 5.34688e+00 Jy')
    assert label_mouseover.icon == 'b'

    # Out-of-bounds shows only cursor info.
    label_mouseover._viewer_mouse_event(spec_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 5500, 'y': 120}})
    assert label_mouseover.as_text() == ('Cursor 5.50000e+03, 1.20000e+02', '', '')
    assert label_mouseover.icon == 'mdi-cursor-default'


def test_specviz2d_smooth(specviz2d_helper, spectrum2d):
    data_label = 'test'
    dc = specviz2d_helper.app.data_collection
    specviz2d_helper.load_data(spectrum_2d=spectrum2d, spectrum_2d_label=data_label)

    gs_plugin = specviz2d_helper.plugins['Gaussian Smooth']

    # The Autocollapsed spectrum is given the label of "Spectrum 1D" by default
    smooth_source_dataset = "Spectrum 1D"
    gs_plugin.dataset = smooth_source_dataset
    test_stddev_level = 10.0
    gs_plugin.stddev = test_stddev_level
    smoothed_spectrum = gs_plugin.smooth(add_data=True)

    assert len(dc) == 3
    assert dc[2].label == f'{smooth_source_dataset} smooth stddev-{test_stddev_level}'
    np.testing.assert_allclose(smoothed_spectrum.spectral_axis.value,
                               spectrum2d.spectral_axis.value)

    # Ensure all marks were created properly (i.e. not in their initialized state)
    # [0,1] is the default (initialization) value for the marks
    marks = specviz2d_helper.app.get_viewer('spectrum-viewer').native_marks
    assert len(marks) == 2

    gp_mark = marks[-1]
    np.testing.assert_allclose(gp_mark.x, smoothed_spectrum.spectral_axis.value)
    np.testing.assert_allclose(gp_mark.y, smoothed_spectrum.flux.value)



from traitlets import Unicode

from jdaviz.core.registries import loader_resolver_registry
from jdaviz.core.loaders.resolvers import BaseResolver
from jdaviz.core.user_api import LoaderUserApi


@loader_resolver_registry('object')
class ObjectResolver(BaseResolver):
    template_file = __file__, "object.vue"
    default_input = 'object'
    requires_api_support = True

    object_repr = Unicode("").tag(sync=True)

    def __init__(self, *args, **kwargs):
        self._object = None
        super().__init__(*args, **kwargs)

    @property
    def user_api(self):
        return LoaderUserApi(self, expose=['object'])

    @property
    def is_valid(self):
        return not isinstance(self.object, str)

    @property
    def object(self):
        return self._object

    @object.setter
    def object(self, obj):
        self._object = obj
        self.object_repr = f"<{obj.__class__.__name__} object>"
        self._update_format_items()

    def __call__(self):
        return self.object



import os
from traitlets import Unicode, observe

from jdaviz.configs.default.plugins.data_tools.file_chooser import FileChooser
from jdaviz.core.registries import loader_resolver_registry
from jdaviz.core.loaders.resolvers import BaseResolver
from jdaviz.core.user_api import LoaderUserApi


__all__ = ['FileResolver']


@loader_resolver_registry('file')
class FileResolver(BaseResolver):
    template_file = __file__, "file.vue"
    default_input = 'filepath'

    filepath = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        start_path = os.environ.get('JDAVIZ_START_DIR', os.path.curdir)
        self._file_upload = FileChooser(start_path)
        self._file_upload.observe(self._on_file_upload_path_changed, names='file_path')
        self.components = {'g-file-import': self._file_upload}
        super().__init__(*args, **kwargs)

    @property
    def user_api(self):
        return LoaderUserApi(self, expose=['filepath'])

    def _on_file_upload_path_changed(self, event):
        self.filepath = self._file_upload.file_path

    @observe('filepath')
    def _on_filepath_changed(self, change):
        if self._file_upload.file_path != change['new']:
            path, file = os.path.split(change['new'])
            self._file_upload._set_form_values(path, file)
        self._update_format_items()

    @property
    def is_valid(self):
        if self._file_upload.file_path is None:
            return False
        return os.path.exists(self._file_upload.file_path) and os.path.isfile(self._file_upload.file_path)  # noqa

    def __call__(self):
        return self._file_upload.file_path



import astropy.units as u
from itertools import combinations
from numpy.testing import assert_allclose
import pytest
from specutils import Spectrum1D

from jdaviz.core.custom_units_and_equivs import (PIX2,
                                                 SPEC_PHOTON_FLUX_DENSITY_UNITS,
                                                 _eqv_pixar_sr)
from jdaviz.core.unit_conversion_utils import (all_flux_unit_conversion_equivs,
                                               check_if_unit_is_per_solid_angle,
                                               combine_flux_and_angle_units,
                                               flux_conversion_general,
                                               handle_squared_flux_unit_conversions,
                                               viewer_flux_conversion_equivalencies)


@pytest.mark.parametrize("unit, is_solid_angle", [
    ('erg / sr', True), ('erg / (deg * deg)', True), ('erg / (deg * arcsec)', True),
    ('erg / (deg**2)', True), ('erg deg**-2', True), ('erg sr^-1', True),
    ('erg / degree', False), ('erg sr^-2', False)])
def test_check_if_unit_is_per_solid_angle(unit, is_solid_angle):
    # test function that tests if a unit string or u.Unit is per solid angle
    assert check_if_unit_is_per_solid_angle(unit) == is_solid_angle

    # turn string into u.Unit, and make sure it also passes
    assert check_if_unit_is_per_solid_angle(u.Unit(unit)) == is_solid_angle


def test_general_flux_conversion():
    """
    This test function tests that the `flux_conversion_general` function is able
    to convert between all available flux units in the unit conversion plugin
    for data in spectral/photon flux density (e.g Jy, erg/s/cm2/A) or surface
    brightness. `flux_conversion_general` handles special cases where flux and
    surface brightness can't be converted directly, so it is used
    in place of astropy.units.to() across the application to handle flux/sb unit
    conversions. This test ensures that all advertised conversions are supported
    and correct.

    Also tests the `handle_squared_flux_unit_conversions` function, which is used
    to convert units for the aperture photometry table.
    """

    # required equivalencies for all flux<>flux, flux<>sb, sb<>sb conversions
    equivalencies = all_flux_unit_conversion_equivs(pixar_sr=4, cube_wave=1*u.nm)

    # first test all conversions between flux units and surface brightness units
    # (per steradian). (just test that it returns without error)
    sr_sbs = combine_flux_and_angle_units(SPEC_PHOTON_FLUX_DENSITY_UNITS, u.sr)
    all_convertable_units_sr = SPEC_PHOTON_FLUX_DENSITY_UNITS + sr_sbs
    for combo in combinations(all_convertable_units_sr, 2):
        original_unit, target_unit = (u.Unit(x) for x in combo)
        converted = flux_conversion_general([1, 2, 3], original_unit,
                                            target_unit, equivalencies)
        assert len(converted) == 3

    # next test all conversions between flux and surface brightness units (per
    # square pixel), omitting the flux<>flux conversions already covered above
    pix_sbs = combine_flux_and_angle_units(SPEC_PHOTON_FLUX_DENSITY_UNITS, PIX2)
    all_convertable_units_pix = SPEC_PHOTON_FLUX_DENSITY_UNITS + pix_sbs
    all_convertable_units_pix = set(all_convertable_units_pix) - set(all_convertable_units_sr)
    for combo in combinations(all_convertable_units_pix, 2):
        original_unit, target_unit = (u.Unit(x) for x in combo)
        converted = flux_conversion_general([1, 2, 3], original_unit,
                                            target_unit, equivalencies)
        assert len(converted) == 3

    # test that a unit combination passed in without the correct equivalency
    # raises the correct error
    msg = 'Could not convert Jy / pix2 to Jy / sr with provided equivalencies.'
    with pytest.raises(u.UnitConversionError, match=msg):
        converted = flux_conversion_general([1, 2, 3], u.Jy / PIX2, u.Jy / u.sr)

    # and finally, numerically verify a subset of possible unit conversion combos
    # a case of each 'type' of conversion is covered here
    units_and_expected = [(u.Jy / u.sr, u.MJy, 4.0e-6),
                          (u.Jy, u.MJy, 1.0e-6),
                          (u.Jy, u.MJy / u.sr, 2.5e-7),
                          (u.Jy, u.MJy / PIX2, 1.0e-6),
                          (u.Jy / PIX2, u.MJy / PIX2, 1.0e-6),
                          (u.MJy, u.erg / (u.s * u.cm**2 * u.AA), 0.29979246),
                          (u.MJy / PIX2, u.erg / (u.s * u.cm**2 * u.AA), 0.29979246),
                          (u.MJy, u.erg / (u.s * u.cm**2 * u.AA * PIX2), 0.29979246),
                          (u.MJy / PIX2, u.erg / (u.s * u.cm**2 * u.AA * PIX2), 0.29979246),
                          (u.MJy / u.sr, u.erg / (u.s * u.cm**2 * u.AA * u.sr), 0.29979246),
                          (u.MJy / u.sr, u.erg / (u.s * u.cm**2 * u.AA), 1.1991698),
                          (u.MJy / u.sr, u.erg / (u.s * u.cm**2 * u.AA), 1.1991698),
                          (u.MJy, u.erg / (u.s * u.cm**2 * u.AA * u.sr), 0.07494811)]

    equivalencies = all_flux_unit_conversion_equivs(pixar_sr=4, cube_wave=1*u.nm)
    for orig, targ, truth in units_and_expected:
        converted_value = flux_conversion_general([1], orig, targ, equivalencies)
        assert_allclose(converted_value[0].value, truth)
        assert converted_value.unit == targ

        # as a bonus, also test the function that converts squared flux units
        # (relevant in aperture photometry)
        sq = handle_squared_flux_unit_conversions(1, orig**2, targ**2, equivalencies)
        assert_allclose(sq.value, truth**2, rtol=1e-06)
        assert sq.unit == targ ** 2


def test_general_flux_conversion_continuted():

    """
    This is effectivley the same test as 'test_general_flux_conversion',
    and is included to retain test coverage of a deprecated function that tested
    these specific cases, as to not remove any test coverage
    since these tests are quite fast.
    """

    # the values to be translated in every test case
    values = [10, 20, 30]

    # unit conversion equivalencies for u.spectral, and flux<>SB for all flux
    # conversions supported. Using PIXAR_SR = 0.1 for flux to SB per steradian,
    # and a spectral axis for u.spectral equivalency. depending on the conversion,
    # all of these equivs might not be required but it is harmless to pass them anyway.
    equivs = all_flux_unit_conversion_equivs(pixar_sr=0.1,
                                             cube_wave=[1, 2, 3] * u.um)

    test_combos = [(u.Jy / u.sr, u.Jy, [1, 2, 3]),
                   (u.Jy, u.Jy / u.sr, [100, 200, 300]),
                   (u.Jy / PIX2, u.Jy, [10, 20, 30]),
                   (u.Jy, u.Jy / PIX2, [10, 20, 30]),
                   (u.Jy / u.sr, u.erg / (u.Angstrom * u.s * u.cm**2 * u.sr),
                   [2.99792458e-12, 1.49896229e-12, 9.99308193e-13]),
                   (u.Jy / PIX2, u.erg / (u.Angstrom * u.s * u.cm**2 * PIX2),
                   [2.99792458e-12, 1.49896229e-12, 9.99308193e-13]),
                   (u.erg / (u.Angstrom * u.s * u.cm**2 * u.sr), u.Jy / u.sr,
                   [3.33564095e+13, 2.66851276e+14, 9.00623057e+14]),
                   (u.erg / (u.Angstrom * u.s * u.cm**2 * PIX2), u.Jy / PIX2,
                   [3.33564095e+13, 2.66851276e+14, 9.00623057e+14]),
                   (u.erg / (u.Angstrom * u.s * u.cm**2 * u.sr), u.Jy / u.sr,
                   [3.33564095e+13, 2.66851276e+14, 9.00623057e+14]),
                   (u.erg / (u.Angstrom * u.s * u.cm**2 * PIX2), u.Jy / PIX2,
                   [3.33564095e+13, 2.66851276e+14, 9.00623057e+14]),
                   (u.MJy, u.ph / (u.s * u.cm**2 * u.Hz * u.sr),
                   [0.00050341, 0.00201365, 0.0045307]),
                   (u.MJy, u.ph / (u.s * u.cm**2 * u.Hz),
                   [5.03411657e-05, 2.01364663e-04, 4.53070491e-04]),
                   (u.ph / (u.s * u.cm**2 * u.Hz * u.sr), u.MJy,
                   [198644.58571489, 198644.58571489, 198644.58571489]),
                   (u.ph / (u.s * u.cm**2 * u.Hz * PIX2), u.MJy,
                   [1986445.85714893, 1986445.85714893, 1986445.85714893]),
                   (u.Jy / u.sr, u.MJy, [1.e-06, 2.e-06, 3.e-06]),
                   (u.MJy / PIX2, u.MJy, [10, 20, 30])
                   ]

    for orig, targ, expected in test_combos:
        converted = flux_conversion_general(values, orig, targ, equivs)
        assert_allclose(converted, expected * targ, rtol=1e-05)

    # some other misc test cases from previous test
    converted = flux_conversion_general(1., u.MJy / u.sr,
                                        u.erg / (u.s * u.cm**2 * u.Hz * u.sr),
                                        equivs)
    assert_allclose(converted, 1.e-17 * u.erg / (u.s * u.cm**2 * u.Hz * u.sr))

    # another old test case
    converted = flux_conversion_general(10., u.MJy / u.sr, u.Jy / u.sr)
    assert_allclose(converted, 10000000. * u.Jy / u.sr)

    # Quantity scalar pixel scale factor
    eqv = _eqv_pixar_sr(0.1 * (u.sr / u.pix))
    assert_allclose(flux_conversion_general(values, u.Jy / u.sr, u.Jy, eqv),
                    [1, 2, 3] * u.Jy)
    assert_allclose(flux_conversion_general(values, u.Jy, u.Jy / u.sr, eqv),
                    [100, 200, 300] * u.Jy / u.sr)

    # values == 2
    assert_allclose(flux_conversion_general([10, 20], u.Jy / u.sr, u.Jy, eqv),
                    [1, 2] * u.Jy)
    assert_allclose(flux_conversion_general([10, 20], u.Jy, u.Jy / u.sr, eqv),
                    [100, 200] * u.Jy / u.sr)

    # array scale factor, same length arrays
    res = flux_conversion_general(values, u.Jy / u.sr, u.Jy,
                                  _eqv_pixar_sr([0.1, 0.2, 0.3]))
    assert_allclose(res, [1., 4., 9.] * u.Jy)

    res = flux_conversion_general(values, u.Jy, u.Jy / u.sr,
                                  _eqv_pixar_sr([0.1, 0.2, 0.3]))
    assert_allclose(res, [100, 100, 100] * u.Jy / u.sr)

    # array + quantity scale factor, same length arrays
    eqv = _eqv_pixar_sr([0.1, 0.2, 0.3] * (u.sr / u.pix))
    assert_allclose(flux_conversion_general(values, u.Jy / u.sr, u.Jy, eqv),
                    [1., 4., 9.] * u.Jy)
    assert_allclose(flux_conversion_general(values, u.Jy, u.Jy / u.sr, eqv),
                    [100, 100, 100] * u.Jy / u.sr)

    # values != 2 but _pixel_scale_factor size mismatch
    with pytest.raises(ValueError, match="operands could not be broadcast together"):
        eqv = _eqv_pixar_sr([0.1, 0.2, 0.3, 0.4])
        flux_conversion_general(values, u.Jy / u.sr, u.Jy, eqv)

    # Other kind of flux conversion unrelated to _pixel_scale_factor.
    # The answer was obtained from synphot unit conversion.
    targ = [2.99792458e-12, 1.49896229e-12, 9.99308193e-13]
    targ *= (u.erg / (u.AA * u.cm * u.cm * u.s))
    assert_allclose(flux_conversion_general(values, u.Jy, targ.unit, equivs), targ)


def test_viewer_flux_conversion():
    """
    This function tests flux/sb conversion specifically for the case of the
    spectrum viewer, where the assumption that 2 values being converted == viewer
    limits is made and the equivalencies list is determined slightly differently
    than in the general case (e.g when spectral axis length is greater that 2
    but 2 values are being converted, spectral_axis[0] is used. also if there is
    a longer list of pixel scale factors but 2 values being converted, the min/max
    of the scale factors are used.).
    """

    # use viewer_flux_conversion_equivalencies to generate list of equivalencies,
    # which requires a Spectrum1D
    spec = Spectrum1D(flux=[1, 2, 3]*u.Jy, spectral_axis=[1, 2, 3]*u.um,
                      meta={'_pixel_scale_factor': [0.1, 0.2, 0.3]})
    viewer_equivs = viewer_flux_conversion_equivalencies([10, 20], spec)

    # test for when scale factor array len > 2 but there are 2 values to be
    # converted, the min/max should be used. min_max = [0.1, 0.3]
    res = flux_conversion_general([10, 20], u.Jy / u.sr, u.Jy,
                                  viewer_equivs)
    assert_allclose(res, [1, 6] * u.Jy)
    res = flux_conversion_general([10, 20], u.Jy, u.Jy / u.sr,
                                  viewer_equivs)
    assert_allclose(res, [100, 66.66666666666667] * u.Jy / u.sr)

    # test converting 2 values when spectral axis for u.spectral,
    # contains 3 items. This will only occur in the viewer limits case, so
    values = [10, 20]
    targ = [2.99792458e-12, 5.99584916e-12]
    targ *= (u.erg / (u.AA * u.cm * u.cm * u.s))  # FLAM
    assert_allclose(flux_conversion_general(values, u.Jy, targ.unit, viewer_equivs),
                    targ)



import warnings

import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import SkyCoord, Angle
from astropy.utils.data import get_pkg_data_filename
from numpy.testing import assert_allclose
from photutils.aperture import CircularAperture, SkyCircularAperture
from regions import (PixCoord, CircleSkyRegion, RectanglePixelRegion, CirclePixelRegion,
                     EllipsePixelRegion, PointSkyRegion, PolygonPixelRegion,
                     CircleAnnulusPixelRegion, CircleAnnulusSkyRegion, Regions)

from jdaviz.configs.imviz.tests.utils import BaseImviz_WCS_NoWCS


class BaseRegionHandler:
    """Test to see if region is loaded.
    Does not check if region is actually at the correct place in display.
    """

    def verify_region_loaded(self, region_label, count=2):
        n = 0
        for layer in self.viewer.state.layers:
            if layer.layer.label == region_label:
                n += 1
                assert layer.visible
        assert n == count


class TestLoadRegions(BaseImviz_WCS_NoWCS, BaseRegionHandler):
    def teardown_method(self, method):
        """Clear all the subsets for the next test method."""
        self.imviz.app.delete_subsets()

    def test_regions_invalid(self):
        # Wrong object
        bad_regions = self.subset_plugin.import_region([self.imviz], return_bad_regions=True)
        assert len(bad_regions) == 1 and bad_regions[0][1] == 'Mask creation failed'

        # Sky region on image without WCS
        sky = SkyCoord(337.51894337, -20.83208305, unit='deg')
        reg = CircleSkyRegion(center=sky, radius=0.0004 * u.deg)
        bad_regions = self.subset_plugin.import_region([reg], refdata_label='no_wcs[SCI,1]',
                                                       return_bad_regions=True)
        assert len(bad_regions) == 1 and bad_regions[0][1] == 'Sky region provided but data has no valid WCS'  # noqa

        reg = SkyCircularAperture(sky, 0.5 * u.arcsec)
        bad_regions = self.subset_plugin.import_region([reg], refdata_label='no_wcs[SCI,1]',
                                                       return_bad_regions=True)
        assert len(bad_regions) == 1 and bad_regions[0][1] == 'Sky region provided but data has no valid WCS'  # noqa

        reg = CircleAnnulusSkyRegion(center=sky, inner_radius=0.0004 * u.deg,
                                     outer_radius=0.0005 * u.deg)
        bad_regions = self.subset_plugin.import_region([reg], refdata_label='no_wcs[SCI,1]',
                                                       return_bad_regions=True)
        assert len(bad_regions) == 1 and bad_regions[0][1] == 'Sky region provided but data has no valid WCS'  # noqa

        # Unsupported functionality
        reg = PointSkyRegion(center=sky)
        bad_regions = self.subset_plugin.import_region(reg, return_bad_regions=True)
        assert len(bad_regions) == 1 and bad_regions[0][1] == 'Failed to load: NotImplementedError()'  # noqa

        # Out-of-bounds masked subset (pix)
        reg = PolygonPixelRegion(vertices=PixCoord(x=[11, 12, 12], y=[11, 11, 12]))
        bad_regions = self.subset_plugin.import_region(reg, return_bad_regions=True)
        assert len(bad_regions) == 1 and bad_regions[0][1] == 'Mask creation failed'

        # Make sure nothing is returned when not requested even on failure
        bad_regions = self.subset_plugin.import_region(reg)
        assert bad_regions is None

        # Make sure nothing is actually loaded
        self.verify_region_loaded('MaskedSubset 1', count=0)
        assert self.imviz.plugins['Subset Tools'].get_regions() == {}

    def test_regions_fully_out_of_bounds(self):
        """Glue ROI will not error when out of bounds."""
        my_reg = CirclePixelRegion(center=PixCoord(x=100, y=100), radius=5)
        bad_regions = self.subset_plugin.import_region([my_reg], return_bad_regions=True)
        assert len(bad_regions) == 0
        self.verify_region_loaded('Subset 1')
        assert len(self.imviz.plugins['Subset Tools'].get_regions()) == 1

    def test_regions_mask(self):
        mask = np.zeros((10, 10), dtype=np.bool_)
        mask[0, 0] = True
        bad_regions = self.subset_plugin.import_region([mask], return_bad_regions=True)
        assert len(bad_regions) == 0
        self.verify_region_loaded('MaskedSubset 1')
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore',
                                    message='Regions skipped: MaskedSubset 1')
            assert self.imviz.plugins['Subset Tools'].get_regions() == {}

        mask[1, 1] = True
        bad_regions = self.subset_plugin.import_region([mask], return_bad_regions=True)
        assert len(bad_regions) == 0
        self.verify_region_loaded('MaskedSubset 2')
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore',
                                    message='Regions skipped: MaskedSubset 1')
            assert self.imviz.plugins['Subset Tools'].get_regions() == {}

        # Also test deletion by label here.
        self.imviz.app.delete_subsets('MaskedSubset 1')
        self.verify_region_loaded('MaskedSubset 1', count=0)

        # Adding another mask will increment from 2 even when 1 is now available.
        mask[2, 2] = True
        bad_regions = self.subset_plugin.import_region([mask], return_bad_regions=True)
        assert len(bad_regions) == 0
        self.verify_region_loaded('MaskedSubset 3')
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore',
                                    message='Regions skipped: MaskedSubset 2, MaskedSubset 3')
            assert self.imviz.plugins['Subset Tools'].get_regions() == {}

        # Deletion of non-existent label raises error
        with pytest.raises(ValueError, match=r'foo not in data collection, can not delete\.'):
            self.imviz.app.delete_subsets('foo')

    def test_regions_pixel(self):
        # A little out-of-bounds should still overlay the overlapped part.
        my_reg = CirclePixelRegion(center=PixCoord(x=6, y=2), radius=5)
        bad_regions = self.subset_plugin.import_region([my_reg], return_bad_regions=True,
                                                       combination_mode='new')
        assert len(bad_regions) == 0
        self.verify_region_loaded('Subset 1')
        st = self.imviz.plugins['Subset Tools']
        assert len(st.get_regions()) == 1
        # test passing in the subset label as well
        assert len(st.get_regions(list_of_subset_labels='Subset 1')) == 1

    def test_regions_sky_has_wcs(self):
        my_reg_pix_1 = CirclePixelRegion(center=PixCoord(x=2.55, y=3.55), radius=1.05)
        my_reg_pix_2 = EllipsePixelRegion(center=PixCoord(x=1.5, y=2.25), width=7.0, height=4.5)
        my_reg_pix_3 = RectanglePixelRegion(center=PixCoord(x=5.0, y=5.0), width=10, height=10)

        sky = SkyCoord(ra=337.5202808, dec=-20.833333059999998, unit='deg')
        # These will become indistinguishable from normal Subset.
        my_reg_sky_1 = CircleSkyRegion(sky, Angle(0.5, u.arcsec))
        my_reg_sky_2 = CircleAnnulusSkyRegion(center=sky, inner_radius=0.0004 * u.deg,
                                              outer_radius=0.0005 * u.deg)
        # Masked subset.
        my_reg_poly_1 = PolygonPixelRegion(vertices=PixCoord(x=[1, 1, 3, 3, 1], y=[1, 3, 3, 1, 1]))

        # Add them all.
        bad_regions = self.subset_plugin.import_region(
            [my_reg_pix_1, my_reg_sky_1, my_reg_sky_2, my_reg_poly_1, my_reg_pix_2, my_reg_pix_3],
            return_bad_regions=True, combination_mode='new')
        assert len(bad_regions) == 0

        # Check regions. We do not check if the translation is correct,
        # that check hopefully is already done in glue-astronomy.
        # Polygon becomes MaskedSubset 1, thus cannot round-trip but made Subset 4 name unavailable.
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore',
                                    message='Regions skipped: MaskedSubset 1')
            subsets = self.imviz.plugins['Subset Tools'].get_regions()
        assert list(subsets.keys()) == ['Subset 1', 'Subset 2', 'Subset 3', 'Subset 5', 'Subset 6'], subsets  # noqa: E501
        assert isinstance(subsets['Subset 1'], CirclePixelRegion)
        assert isinstance(subsets['Subset 2'], CirclePixelRegion)
        assert isinstance(subsets['Subset 3'], CircleAnnulusPixelRegion)
        assert isinstance(subsets['Subset 5'], EllipsePixelRegion)
        assert isinstance(subsets['Subset 6'], RectanglePixelRegion)

        # Polygon becomes MaskedSubset 1.
        self.verify_region_loaded('MaskedSubset 1')

    def test_regions_annulus_from_load_data(self):
        # This file actually will load 2 annuli
        regfile = get_pkg_data_filename('data/ds9_annulus_01.reg')
        self.imviz.load_data(regfile)
        assert len(self.imviz.app.data_collection) == 2  # Make sure not loaded as data

        # Test data is set up such that 1 pixel is 1 arcsec.
        subset_radii = {"Subset 1": [0.5, 1], "Subset 2": [1, 3]}

        subsets = self.imviz.plugins['Subset Tools'].get_regions()
        subset_names = sorted(subsets.keys())
        assert subset_names == ['Subset 1', 'Subset 2']
        for n in subset_names:
            assert isinstance(subsets[n], CircleAnnulusPixelRegion)
            assert_allclose(subsets[n].center.xy, 4.5, rtol=5e-6)
            assert_allclose([subsets[n].inner_radius, subsets[n].outer_radius], subset_radii[n])

    def test_photutils_pixel(self):
        my_aper = CircularAperture((5, 5), r=2)
        bad_regions = self.subset_plugin.import_region([my_aper], return_bad_regions=True)
        assert len(bad_regions) == 0
        self.verify_region_loaded('Subset 1')
        assert len(self.imviz.plugins['Subset Tools'].get_regions()) == 1

    def test_photutils_sky_has_wcs(self):
        sky = SkyCoord(ra=337.5202808, dec=-20.833333059999998, unit='deg')
        my_aper_sky = SkyCircularAperture(sky, 0.5 * u.arcsec)
        bad_regions = self.subset_plugin.import_region([my_aper_sky], return_bad_regions=True)
        assert len(bad_regions) == 0
        self.verify_region_loaded('Subset 1')
        assert len(self.imviz.plugins['Subset Tools'].get_regions()) == 1


class TestLoadRegionsFromFile(BaseRegionHandler):

    def setup_class(self):
        self.region_file = get_pkg_data_filename(
            'data/ds9.fits.reg', package='regions.io.ds9.tests')
        self.arr = np.ones((1024, 1024))
        self.raw_regions = Regions.read(self.region_file, format='ds9')

    def test_ds9_load_all(self, imviz_helper):
        with pytest.raises(ValueError, match="Cannot load regions without data"):
            imviz_helper.load_data(self.region_file)

        self.viewer = imviz_helper.default_viewer._obj
        imviz_helper.load_data(self.arr, data_label='my_image')
        bad_regions = imviz_helper.plugins['Subset Tools'].import_region(
            self.region_file, combination_mode='new', return_bad_regions=True)
        assert len(bad_regions) == 1

        # Will load 8/9 and 7 of that become ROIs.
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore',
                                    message='Regions skipped: MaskedSubset 1')
            subsets = imviz_helper.plugins['Subset Tools'].get_regions()
        assert list(subsets.keys()) == ['Subset 1', 'Subset 2', 'Subset 3',
                                        'Subset 4', 'Subset 5', 'Subset 6', 'Subset 7'], subsets

        # The other 1 is MaskedSubset
        self.verify_region_loaded('MaskedSubset 1', count=1)

    def test_ds9_load_two_good(self, imviz_helper):
        self.viewer = imviz_helper.default_viewer._obj
        imviz_helper.load_data(self.arr, data_label='my_image')
        bad_regions = imviz_helper.plugins['Subset Tools'].import_region(
            self.region_file, combination_mode='new', max_num_regions=2, return_bad_regions=True)
        assert len(bad_regions) == 0
        subsets = imviz_helper.plugins['Subset Tools'].get_regions()
        assert list(subsets.keys()) == ['Subset 1', 'Subset 2'], subsets
        self.verify_region_loaded('MaskedSubset 1', count=0)

    def test_ds9_load_one_bad(self, imviz_helper):
        self.viewer = imviz_helper.default_viewer._obj
        imviz_helper.load_data(self.arr, data_label='my_image')
        bad_regions = imviz_helper.plugins['Subset Tools'].import_region(
            self.raw_regions[6], return_bad_regions=True)
        assert len(bad_regions) == 1
        assert imviz_helper.plugins['Subset Tools'].get_regions() == {}
        self.verify_region_loaded('MaskedSubset 1', count=0)

    def test_ds9_load_one_good_one_bad(self, imviz_helper):
        self.viewer = imviz_helper.default_viewer._obj
        imviz_helper.load_data(self.arr, data_label='my_image')
        bad_regions = imviz_helper.plugins['Subset Tools'].import_region(
            [self.raw_regions[3], self.raw_regions[6]], return_bad_regions=True)
        assert len(bad_regions) == 1

        subsets = imviz_helper.plugins['Subset Tools'].get_regions()
        assert list(subsets.keys()) == ['Subset 1'], subsets
        self.verify_region_loaded('MaskedSubset 1', count=0)


class TestGetRegions(BaseImviz_WCS_NoWCS):
    def test_annulus(self):
        self.imviz.plugins['Subset Tools'].import_region([
            CirclePixelRegion(center=PixCoord(x=4.5, y=4.5), radius=4.5),  # Outer circle
            CirclePixelRegion(center=PixCoord(x=4.5, y=4.5), radius=2.5),  # Inner circle
        ], combination_mode="new")

        # At this point, there should be two normal circles.
        subsets = self.imviz.plugins['Subset Tools'].get_regions()
        assert list(subsets.keys()) == ['Subset 1', 'Subset 2'], subsets
        assert isinstance(subsets['Subset 1'], CirclePixelRegion)
        assert isinstance(subsets['Subset 2'], CirclePixelRegion)
        assert subsets['Subset 1'].center == PixCoord(4.5, 4.5)
        assert subsets['Subset 2'].center == PixCoord(4.5, 4.5)
        # ensure agreement between app.get_subsets and subset_tools.get_regions
        ss = self.imviz.app.get_subsets()
        assert ss['Subset 1'][0]['region'] == subsets['Subset 1']
        assert ss['Subset 2'][0]['region'] == subsets['Subset 2']

        # Create a third subset that is an annulus.
        self.imviz.plugins['Subset Tools'].combination_mode = "new"
        subset_groups = self.imviz.app.data_collection.subset_groups
        new_subset = subset_groups[0].subset_state & ~subset_groups[1].subset_state
        self.viewer.apply_subset_state(new_subset)

        subsets = self.imviz.plugins['Subset Tools'].get_regions()
        assert len(self.imviz.app.data_collection.subset_groups) == 3
        assert list(subsets.keys()) == ['Subset 1', 'Subset 2', 'Subset 3'], subsets
        assert isinstance(subsets['Subset 1'], CirclePixelRegion)
        assert isinstance(subsets['Subset 2'], CirclePixelRegion)
        assert isinstance(subsets['Subset 3'], CircleAnnulusPixelRegion)
        # and check the new retrieved region
        assert subsets['Subset 3'].center == PixCoord(4.5, 4.5)

        # Clear the regions for next test.
        self.imviz.app.delete_subsets()
        assert len(self.imviz.app.data_collection.subset_groups) == 0



"""This file contains helper function related to configuration handling."""

import copy
import os
import pathlib
import yaml

__all__ = ['read_configuration', 'get_configuration', 'list_configurations']


def read_configuration(path=None):
    """Loads a configuration from a YAML file.

    Parameters
    ----------
    path : str, optional
        Path to the configuration file to be loaded. If None, loads the
        default configuration.

    Returns
    -------
    config : dict
        A dictionary object.
    """
    if not isinstance(path, (str, type(None))):
        raise ValueError('path must be a string')

    # Parse the default configuration file
    default_path = pathlib.Path(__file__).resolve().parent.parent / "configs"

    if path is None or path == 'default':
        path = default_path / "default" / "default.yaml"
    elif path == 'cubeviz':
        path = default_path / "cubeviz" / "cubeviz.yaml"
    elif path == 'specviz':
        path = default_path / "specviz" / "specviz.yaml"
    elif path == 'mosviz':
        path = default_path / "mosviz" / "mosviz.yaml"
    elif path == 'specviz2d':
        path = default_path / "specviz2d" / "specviz2d.yaml"
    elif path == 'imviz':
        path = default_path / "imviz" / "imviz.yaml"
    elif path == 'rampviz':
        path = default_path / "rampviz" / "rampviz.yaml"
    elif not os.path.isfile(path):
        raise ValueError("Configuration must be path to a .yaml file.")

    with open(path, 'r') as f:
        config = yaml.safe_load(f)

    return config


def get_configuration(path=None, section=None, config=None):
    """Retrieve a copy of a specified configuration.

    Returns a copy of a configuration specification.  If ``path``
    is not specified, then returns a copy of the current application
    configuration if ``config`` is specified.

    Parameters
    ----------
    path : str, optional
        path to the configuration file to be retrieved.
    section : str, optional
        A section of the configuration to retrieve
    config : dict, optional
        An existing configuration dictionary

    Returns
    -------
    cfg : dict
        A configuration specification dictionary

    """
    # read the YAML configuration for the given path
    if path:
        config = read_configuration(path=path)
    else:
        if config is None:
            raise ValueError('Either a path or a pre-existing config must be specified')

    cfg = copy.deepcopy(config)

    # return only a section if requested
    if section:
        return cfg.get(section, None)
    return cfg


def list_configurations():
    """Get a list of pre-built configurations."""

    path = pathlib.Path(__file__).resolve().parent.parent / "configs"
    return [i.stem for i in path.rglob('*.yaml')]



import os
import pathlib
from collections import defaultdict

import numpy as np
from astropy.io.registry import IORegistryError
from astropy.nddata import StdDevUncertainty
from astropy.io import fits
from specutils import Spectrum1D, SpectrumList, SpectrumCollection

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import data_parser_registry
from jdaviz.utils import standardize_metadata, download_uri_to_path


__all__ = ["specviz_spectrum1d_parser"]


@data_parser_registry("specviz-spectrum1d-parser")
def specviz_spectrum1d_parser(app, data, data_label=None, format=None, show_in_viewer=True,
                              concat_by_file=False, cache=None, local_path=os.curdir, timeout=None,
                              load_as_list=False):
    """
    Loads a data file or `~specutils.Spectrum1D` object into Specviz.

    Parameters
    ----------
    data : str, `~specutils.Spectrum1D`, or `~specutils.SpectrumList`
        Spectrum1D, SpectrumList, or path to compatible data file.
    data_label : str
        The Glue data label found in the ``DataCollection``.
    format : str
        Loader format specification used to indicate data format in
        `~specutils.Spectrum1D.read` io method.
    concat_by_file : bool
        If True and there is more than one available extension, concatenate
        the extensions within each spectrum file passed to the parser and
        add a concatenated spectrum to the data collection.
    cache : None, bool, or str
        Cache the downloaded file if the data are retrieved by a query
        to a URL or URI.
    local_path : str, optional
        Cache remote files to this path. This is only used if data is
        requested from `astroquery.mast`.
    timeout : float, optional
        If downloading from a remote URI, set the timeout limit for
        remote requests in seconds (passed to
        `~astropy.utils.data.download_file` or
        `~astroquery.mast.Conf.timeout`).
    load_as_list : bool, optional
        Force the parser to load the input file with the `~specutils.SpectrumList` read function
        instead of `~specutils.Spectrum1D`.
    """

    spectrum_viewer_reference_name = app._jdaviz_helper._default_spectrum_viewer_reference_name
    # If no data label is assigned, give it a unique name
    if not data_label:
        data_label = app.return_data_label(data, alt_name="specviz_data")
    # Still need to standardize the label
    elif isinstance(data_label, (list, tuple)):
        data_label = [app.return_data_label(label, alt_name="specviz_data") for label in data_label]  # noqa
    else:
        data_label = app.return_data_label(data_label, alt_name="specviz_data")

    if isinstance(data, SpectrumCollection):
        raise TypeError("SpectrumCollection detected."
                        " Please provide a Spectrum1D or SpectrumList")
    elif isinstance(data, Spectrum1D):
        # Handle the possibility of 2D spectra by splitting into separate spectra
        if data.flux.ndim == 1:
            data_label = [data_label]
            data = [data]
        elif data.flux.ndim == 2:
            data = split_spectrum_with_2D_flux_array(data)
            data_label = [f"{data_label} [{i}]" for i in range(len(data))]
    # No special processing is needed in this case, but we include it for completeness
    elif isinstance(data, SpectrumList):
        pass
    elif isinstance(data, list):
        # special processing for HDUList
        if isinstance(data, fits.HDUList):
            data = [Spectrum1D.read(data)]
            data_label = [data_label]
        else:
            # list treated as SpectrumList if not an HDUList
            data = SpectrumList.read(data, format=format)
    else:
        # try parsing file_obj as a URI/URL:
        data = download_uri_to_path(data, cache=cache, local_path=local_path, timeout=timeout)

        path = pathlib.Path(data)

        if path.is_dir() or load_as_list:
            data = SpectrumList.read(str(path), format=format)
            if data == []:
                raise ValueError(f"`specutils.SpectrumList.read('{str(path)}')` "
                                 "returned an empty list")
        elif path.is_file():
            try:
                data = Spectrum1D.read(str(path), format=format)
                if data.flux.ndim == 2:
                    data = split_spectrum_with_2D_flux_array(data)
                else:
                    data = [data]
                    data_label = [app.return_data_label(data_label, alt_name="specviz_data")]

            except IORegistryError:
                # Multi-extension files may throw a registry error
                data = SpectrumList.read(str(path), format=format)
        else:
            raise FileNotFoundError("No such file: " + str(path))

    # step through SpectrumList and convert any 2D spectra to 1D spectra
    if isinstance(data, SpectrumList):
        new_data = []
        for spec in data:
            if spec.flux.ndim == 2:
                new_data.extend(split_spectrum_with_2D_flux_array(spec))
            else:
                new_data.append(spec)
        data = SpectrumList(new_data)

        if not isinstance(data_label, (list, tuple)):
            data_label = [f"{data_label} [{i}]" for i in range(len(data))]
        elif len(data_label) != len(data):
            raise ValueError(f"Length of data labels list ({len(data_label)}) is different"
                             f" than length of list of data ({len(data)})")

        # these are used to build a combined spectrum with all
        # input spectra included (taken from https://github.com/spacetelescope/
        # dat_pyinthesky/blob/main/jdat_notebooks/MRS_Mstar_analysis/
        # JWST_Mstar_dataAnalysis_analysis.ipynb)
        wlallorig = []  # to collect wavelengths
        fnuallorig = []  # fluxes
        dfnuallorig = []  # and uncertanties (if present)

        for spec in data:
            for wlind in range(len(spec.spectral_axis)):
                wlallorig.append(spec.spectral_axis[wlind].value)
                fnuallorig.append(spec.flux[wlind].value)

                # because some spec in the list might have uncertainties and
                # others may not, if uncert is None, set to list of NaN. later,
                # if the concatenated list of uncertanties is all nan (meaning
                # they were all nan to begin with, or all None), it will be set
                # to None on the final Spectrum1D
                if spec.uncertainty is not None and spec.uncertainty[wlind] is not None:
                    dfnuallorig.append(spec.uncertainty[wlind].array)
                else:
                    dfnuallorig.append(np.nan)

    # if the entire uncert. array is Nan and the data is not, model fitting won't
    # work (more generally, if uncert[i] is nan/inf and flux[i] is not, fitting will
    # fail, but just deal with the all nan case here since it is straightforward).
    # set uncerts. to None if they are all nan/inf, and display a warning message.
    set_nans_to_none = False
    if isinstance(data, SpectrumList):
        uncerts = dfnuallorig  # alias these for clarity later on
        if uncerts is not None and not np.any(uncerts):
            uncerts = None
            set_nans_to_none = True
    else:
        if data[0].uncertainty is not None:
            uncerts_finite = np.isfinite(data[0].uncertainty.array)
            if not np.any(uncerts_finite):
                data[0].uncertainty = None
                set_nans_to_none = True

    if set_nans_to_none:
        # alert user that we have changed their all-nan uncertainty array to None
        msg = 'All uncertainties are nonfinite, replacing with uncertainty=None.'
        app.hub.broadcast(SnackbarMessage(msg, color="warning", sender=app))

    with app.data_collection.delay_link_manager_update():
        for i, spec in enumerate(data):
            # note: if SpectrumList, this is just going to be the last unit when
            # combined in the next block. should put a check here to make sure
            # units are all the same or collect them in a list?
            wave_units = spec.spectral_axis.unit
            flux_units = spec.flux.unit

            # Make metadata layout conform with other viz.
            spec.meta = standardize_metadata(spec.meta)

            app.add_data(spec, data_label[i])

            # handle display, with the SpectrumList special case in mind.
            if i == 0 and show_in_viewer:
                app.add_data_to_viewer(spectrum_viewer_reference_name, data_label[i])

        if concat_by_file and isinstance(data, SpectrumList):
            # If >1 spectra in the list were opened from the same FITS file,
            # group them by their original FITS filenames and add their combined
            # 1D spectrum to the DataCollection.
            unique_files = group_spectra_by_filename(app.data_collection)
            for filename, datasets in unique_files.items():
                if len(datasets) > 1:
                    spec = combine_lists_to_1d_spectrum(wlallorig, fnuallorig,
                                                        dfnuallorig, wave_units,
                                                        flux_units)

                    # Make metadata layout conform with other viz.
                    spec.meta = standardize_metadata(spec.meta)

                    label = f"Combined [{filename}]"
                    app.add_data(spec, label)

                    if show_in_viewer:
                        app.add_data_to_viewer(spectrum_viewer_reference_name, label)


def group_spectra_by_filename(datasets):
    """
    Organize the elements of a `~glue.core.data_collection.DataCollection` by the name of
    the FITS file they came from, if available via the header card "FILENAME".

    Parameters
    ----------
    datasets : `~glue.core.data_collection.DataCollection`
        Collection of datasets.

    Returns
    -------
    spectra_to_combine : dict
        Datasets of spectra organized by their filename.
    """
    spectra_to_combine = defaultdict(list)

    for data in datasets:
        filename = data.meta.get("FILENAME")
        spectra_to_combine[filename].append(data)

    return spectra_to_combine


def combine_lists_to_1d_spectrum(wl, fnu, dfnu, wave_units, flux_units):
    """
    Combine lists of 1D spectra into a composite `~specutils.Spectrum1D` object.

    Parameters
    ----------
    wl : list of `~astropy.units.Quantity`s
        Wavelength in each spectral channel
    fnu : list of `~astropy.units.Quantity`s
        Flux in each spectral channel
    dfnu : list of `~astropy.units.Quantity`s or None
        Uncertainty on each flux

    Returns
    -------
    spec : `~specutils.Spectrum1D`
        Composite 1D spectrum.
    """
    wlallarr = np.array(wl)
    fnuallarr = np.array(fnu)
    srtind = np.argsort(wlallarr)
    if dfnu is not None:
        dfnuallarr = np.array(dfnu)
        fnuallerr = dfnuallarr[srtind]
    wlall = wlallarr[srtind]
    fnuall = fnuallarr[srtind]

    # units are not being handled properly yet.
    if dfnu is not None:
        unc = StdDevUncertainty(fnuallerr * flux_units)
    else:
        unc = None

    spec = Spectrum1D(flux=fnuall * flux_units, spectral_axis=wlall * wave_units,
                      uncertainty=unc)
    return spec


def split_spectrum_with_2D_flux_array(data):
    """
    Helper function to split Spectrum1D of 2D flux to a SpectrumList of nD objects.

    Parameters
    ----------
    data : `~specutils.Spectrum1D`
        Spectrum with 2D flux array

    Returns
    -------
    new_data : `~specutils.SpectrumList`
        List of unpacked spectra.
    """
    new_data = []
    for i in range(data.flux.shape[0]):
        unc = None
        mask = None
        if data.uncertainty is not None:
            unc = data.uncertainty[i, :]
        if data.mask is not None:
            mask = data.mask[i, :]
        new_data.append(Spectrum1D(flux=data.flux[i, :], spectral_axis=data.spectral_axis,
                                   uncertainty=unc, mask=mask, meta=data.meta))

    return new_data



import os
import numpy as np
import pytest
from astropy import units as u
from astropy.io import fits
from astropy.nddata import NDData, StdDevUncertainty
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.data import download_file
from astropy.wcs import WCS
from gwcs import WCS as GWCS
from numpy.testing import assert_allclose, assert_array_equal
from regions import CirclePixelRegion, EllipsePixelRegion, PixCoord, RectanglePixelRegion
from skimage.io import imsave
from stdatamodels import asdf_in_fits

from jdaviz.configs.imviz.helper import split_filename_with_fits_ext
from jdaviz.configs.imviz.plugins.parsers import (
    parse_data, _validate_fits_image2d, _validate_bunit, _parse_image, HAS_ROMAN_DATAMODELS)
from jdaviz.core.custom_units_and_equivs import PIX2


@pytest.mark.parametrize(
    ('filename', 'ans'),
    [('/path/to/cache/contents', ['/path/to/cache/contents', None, 'contents']),
     ('file://path/image.fits[SCI]', ['file://path/image.fits', 'SCI', 'image']),
     ('image.fits[dq,2]', ['image.fits', ('dq', 2), 'image']),
     ('/path/to/image.fits', ['/path/to/image.fits', None, 'image']),
     ('/path/to/image.fits[*]', ['/path/to/image.fits', '*', 'image']),
     ('../image.fits.gz[1]', ['../image.fits.gz', 1, 'image.fits'])])
def test_filename_split(filename, ans):
    filepath, ext, data_label = split_filename_with_fits_ext(filename)
    assert filepath == ans[0]
    if ans[1] is None:
        assert ext is None
    else:
        assert ext == ans[1]
    assert data_label == ans[2]


def test_validate_fits_image2d():
    # Not 2D image
    hdu = fits.ImageHDU([0, 0])
    assert not _validate_fits_image2d(hdu, raise_error=False)
    with pytest.raises(ValueError, match='Imviz cannot load this HDU'):
        _validate_fits_image2d(hdu)

    # 2D Image
    hdu = fits.ImageHDU([[0, 0], [0, 0]])
    assert _validate_fits_image2d(hdu)


def test_validate_bunit():
    with pytest.raises(ValueError):
        _validate_bunit('NOT_A_UNIT')

    assert not _validate_bunit('Mjy-sr', raise_error=False)  # Close but not quite
    assert _validate_bunit('MJy/sr') == 'MJy/sr'
    assert _validate_bunit('ELECTRONS/S') == 'electron/s'
    assert _validate_bunit('ELECTRONS') == 'electron'


class TestParseImage:
    def setup_class(self):
        self.jwst_asdf_url_1 = 'https://data.science.stsci.edu/redirect/JWST/jwst-data_analysis_tools/imviz_test_data/jw00042001001_01101_00001_nrcb5_cal.fits'  # noqa: E501
        self.jwst_asdf_url_2 = 'https://stsci.box.com/shared/static/d5k9z5j05dgfv6ljgie483w21kmpevni.fits'  # noqa: E501

    def test_no_data_label(self):
        with pytest.raises(ValueError, match='app is None'):
            _parse_image(None, None, None, False)

    def test_hdulist_no_image(self, imviz_helper):
        hdulist = fits.HDUList([fits.PrimaryHDU()])
        with pytest.raises(ValueError, match='does not have any FITS image'):
            parse_data(imviz_helper.app, hdulist)

    @pytest.mark.parametrize('some_obj', (WCS(), [[1, 2], [3, 4]]))
    def test_invalid_file_obj(self, imviz_helper, some_obj):
        with pytest.raises(NotImplementedError, match='Imviz does not support'):
            parse_data(imviz_helper.app, some_obj)

    def test_parse_numpy_array_1d_2d(self, imviz_helper):
        with pytest.raises(ValueError, match='Imviz cannot load this array with ndim=1'):
            parse_data(imviz_helper.app, np.zeros(2))

        # Passing in data_label keyword as posarg.
        imviz_helper.load_data(np.zeros((2, 2)), 'some_array', show_in_viewer=False)
        data = imviz_helper.app.data_collection[0]
        comp = data.get_component('DATA')
        assert data.label == 'some_array'
        assert data.shape == (2, 2)
        assert comp.data.shape == (2, 2)

    @pytest.mark.parametrize('manual_loop', [False, True])
    def test_parse_numpy_array_3d(self, imviz_helper, manual_loop):
        # This data has values in axis=0 that correspond to axis num.
        n_slices = 5
        slice_shape = (2, 3)
        arr = np.stack([np.zeros(slice_shape) + i for i in range(n_slices)])
        data_label = 'my_slices'

        if not manual_loop:
            # We use higher level load_data() here to make sure linking does not crash.
            imviz_helper.load_data(arr, data_label=data_label)
        else:
            with imviz_helper.batch_load():
                for i in range(n_slices):
                    imviz_helper.load_data(arr[i, :, :], data_label=data_label)

        assert len(imviz_helper.app.data_collection) == n_slices
        assert len(imviz_helper.app.data_collection.links) == 8

        for i in range(n_slices):
            data = imviz_helper.app.data_collection[i]
            comp = data.get_component('DATA')
            assert data.label == (f'my_slices ({i})' if i > 0 else 'my_slices')
            assert data.shape == slice_shape
            assert_array_equal(comp.data, i)

    def test_parse_numpy_array_3d_too_many(self, imviz_helper):
        with pytest.warns(UserWarning, match='16 or more 3D slices found'):
            imviz_helper.load_data(np.ones((17, 5, 5)))

        assert len(imviz_helper.app.data_collection) == 16
        assert imviz_helper.app.data_collection[0].shape == (5, 5)
        assert imviz_helper.app.data_collection[15].shape == (5, 5)

    def test_parse_numpy_array_4d(self, imviz_helper):
        # Check logic is in higher level method.
        imviz_helper.load_data(np.ones((1, 2, 5, 5)))
        assert len(imviz_helper.app.data_collection) == 2
        assert imviz_helper.app.data_collection[0].shape == (5, 5)
        assert imviz_helper.app.data_collection[1].shape == (5, 5)

        with pytest.raises(ValueError, match='cannot load this array with ndim'):
            imviz_helper.load_data(np.ones((2, 2, 5, 5)))

    def test_parse_nddata_simple(self, imviz_helper):
        with pytest.raises(ValueError, match='Imviz cannot load this NDData with ndim=1'):
            parse_data(imviz_helper.app, NDData([1, 2, 3, 4]))

        ndd = NDData([[1, 2], [3, 4]])
        imviz_helper.load_data(ndd, data_label='some_data', show_in_viewer=False)
        data = imviz_helper.app.data_collection[0]
        comp = data.get_component('DATA')
        assert data.label == 'some_data[DATA]'
        assert data.shape == (2, 2)
        assert comp.data.shape == (2, 2)
        assert len(imviz_helper.app.data_collection) == 1

    @pytest.mark.parametrize(
        ('ndd', 'attributes'),
        [(NDData([[1, 2], [3, 4]], mask=[[True, False], [False, False]]),
          ['DATA', 'MASK']),
         (NDData([[1, 2], [3, 4]], uncertainty=StdDevUncertainty([[0.1, 0.2], [0.3, 0.4]])),
          ['DATA', 'UNCERTAINTY'])])
    def test_parse_nddata_with_one_only(self, imviz_helper, ndd, attributes):
        imviz_helper.load_data(ndd, data_label='some_data', show_in_viewer=False)
        for i, attrib in enumerate(attributes):
            data = imviz_helper.app.data_collection[i]
            comp = data.get_component(attrib)
            assert data.label == f'some_data[{attrib}]'
            assert data.shape == (2, 2)
            assert comp.data.shape == (2, 2)
        assert len(imviz_helper.app.data_collection) == 2

    def test_parse_nddata_with_everything(self, imviz_helper):
        ndd = NDData([[1, 2], [3, 4]], mask=[[True, False], [False, False]],
                     uncertainty=StdDevUncertainty([[0.1, 0.2], [0.3, 0.4]]),
                     unit=u.MJy/u.sr, wcs=WCS(naxis=2), meta={'name': 'my_ndd'})
        imviz_helper.load_data(ndd, data_label='some_data', show_in_viewer=False)
        for i, attrib in enumerate(['DATA', 'MASK', 'UNCERTAINTY']):
            data = imviz_helper.app.data_collection[i]
            comp = data.get_component(attrib)
            assert data.label == f'some_data[{attrib}]'
            assert data.shape == (2, 2)
            assert data.meta['name'] == 'my_ndd'
            assert isinstance(data.coords, WCS)
            assert comp.data.shape == (2, 2)
            if attrib == 'MASK':
                assert comp.units == ''
            else:
                assert comp.units == 'MJy / sr'
        assert len(imviz_helper.app.data_collection) == 3

    @pytest.mark.filterwarnings('ignore:.* is a low contrast image')
    @pytest.mark.parametrize('format', ('jpg', 'png'))
    def test_parse_rgba(self, imviz_helper, tmp_path, format):
        if format == 'png':  # Cross-test PNG with RGBA
            a = np.zeros((10, 10, 4), dtype='uint8')
        else:  # Cross-test JPG with RGB only
            a = np.zeros((10, 10, 3), dtype='uint8')

        filename = str(tmp_path / f'myimage.{format}')
        imsave(filename, a)

        imviz_helper.load_data(filename, show_in_viewer=False)
        data = imviz_helper.app.data_collection[0]
        assert data.label == 'myimage'
        assert data.shape == (10, 10)

    def test_filelist(self, imviz_helper, tmp_path):
        flist = []

        # Generate some files to parse.
        for i in range(2):
            fpath = tmp_path / f'myfits_{i}.fits'
            flist.append(str(fpath))
            hdu = fits.PrimaryHDU(np.zeros((2, 2)) + i)
            hdu.header['foo'] = 'bar'
            hdu.writeto(fpath, overwrite=True)

        flist = ','.join(flist)
        imviz_helper.load_data(flist, show_in_viewer=False)

        for i in range(2):
            data = imviz_helper.app.data_collection[i]
            comp = data.get_component('PRIMARY,1')
            assert data.label == f'myfits_{i}[PRIMARY,1]'
            assert data.shape == (2, 2)
            assert data.meta['FOO'] == 'bar'
            np.testing.assert_allclose(comp.data.mean(), i)

        with pytest.raises(ValueError, match='Do not manually overwrite data_label'):
            imviz_helper.load_data(flist, data_label='foo', show_in_viewer=False)

    def test_parse_asdf_in_fits_4d(self, imviz_helper, tmp_path):
        hdulist = fits.HDUList([
            fits.PrimaryHDU(),
            fits.ImageHDU(np.zeros((1, 2, 5, 5)), name='SCI')])
        tree = {'data': {'data': hdulist['SCI'].data}}
        filename = str(tmp_path / 'myasdf.fits')
        asdf_in_fits.write(filename, tree, hdulist=hdulist, overwrite=True)

        with pytest.raises(ValueError, match='Imviz cannot load this HDU'):
            parse_data(imviz_helper.app, filename)

    @pytest.mark.remote_data
    def test_parse_jwst_nircam_level2(self, imviz_helper):
        # Default behavior: Science image
        with pytest.warns(UserWarning, match='You may be querying for a remote file'):
            # if you don't pass a `cache` value, a warning should be raised:
            imviz_helper.load_data(self.jwst_asdf_url_1, timeout=100)

        data = imviz_helper.app.data_collection[0]
        comp = data.get_component('DATA')
        expected_label = os.path.splitext(os.path.basename(self.jwst_asdf_url_1))[0] + '[DATA]'
        assert data.label == expected_label
        assert data.shape == (2048, 2048)
        assert isinstance(data.coords, GWCS)
        assert comp.units == 'MJy/sr'
        assert comp.data.shape == (2048, 2048)

        # --- Since download is expensive, we attach GWCS-specific tests here. ---

        # Ensure region supports GWCS. Also see test_regions.py
        imviz_helper.plugins['Subset Tools'].import_region([
            CirclePixelRegion(center=PixCoord(x=970.95, y=1116.05), radius=5.95),
            RectanglePixelRegion(center=PixCoord(x=995.0, y=1082.5), width=26, height=11)
        ], combination_mode="new")

        subsets = imviz_helper.plugins['Subset Tools'].get_regions()
        assert list(subsets.keys()) == ['Subset 1', 'Subset 2'], subsets
        # check that retrieved subsets-as-regions from subset plugin match what was loaded.
        assert isinstance(subsets['Subset 1'], CirclePixelRegion)
        assert subsets['Subset 1'].center == PixCoord(970.95, 1116.05)
        assert isinstance(subsets['Subset 2'], RectanglePixelRegion)
        assert subsets['Subset 2'].center == PixCoord(995.0, 1082.5)
        # ensure agreement between app.get_subsets and subset_tools.get_regions
        ss = imviz_helper.app.get_subsets()
        assert ss['Subset 1'][0]['region'] == subsets['Subset 1']
        assert ss['Subset 2'][0]['region'] == subsets['Subset 2']

        # Test simple aperture photometry plugin.
        phot_plugin = imviz_helper.app.get_tray_item_from_name('imviz-aper-phot-simple')
        phot_plugin.data_selected = 'contents[DATA]'
        phot_plugin.aperture_selected = 'Subset 1'
        assert phot_plugin.aperture.selected_validity.get('is_aperture')
        assert_allclose(phot_plugin.background_value, 0)
        phot_plugin.background_selected = 'Subset 2'
        assert phot_plugin.aperture.selected_validity.get('is_aperture')
        assert_allclose(phot_plugin.background_value, 0.1741226315498352)  # Subset 2 median
        # NOTE: jwst.datamodels.find_fits_keyword("PHOTMJSR")
        phot_plugin.counts_factor = (data.meta['photometry']['conversion_megajanskys'] /
                                     data.meta['exposure']['exposure_time'])
        assert_allclose(phot_plugin.counts_factor, 0.0036385915646798953)
        assert_allclose(phot_plugin.flux_scaling, 0.003631)
        phot_plugin.vue_do_aper_phot()
        tbl = imviz_helper.plugins['Aperture Photometry'].export_table()
        assert_quantity_allclose(tbl[0]['xcenter'], 970.95 * u.pix)
        assert_quantity_allclose(tbl[0]['ycenter'], 1116.05 * u.pix)
        sky = tbl[0]['sky_center']
        assert_allclose(sky.ra.deg, 80.48419863)
        assert_allclose(sky.dec.deg, -69.494608)
        data_unit = u.MJy / u.sr
        assert_quantity_allclose(tbl[0]['background'], 0.1741226315498352 * data_unit)
        assert_quantity_allclose(tbl[0]['sum'], 4.486487e-11 * u.MJy, rtol=1e-6)
        assert_quantity_allclose(tbl[0]['sum_aper_area'], 111.220234 * PIX2)
        assert_quantity_allclose(tbl[0]['pixarea_tot'], 9.33677e-14 * u.sr, rtol=1e-6)
        assert_quantity_allclose(tbl[0]['aperture_sum_counts'], 132061.576643 * u.count, rtol=1e-6)
        assert_quantity_allclose(tbl[0]['aperture_sum_counts_err'], 363.402775 * u.count)
        assert_quantity_allclose(tbl[0]['counts_fac'], 0.0036385915646798953 * (data_unit / u.ct))
        assert_quantity_allclose(tbl[0]['aperture_sum_mag'], 19.770299 * u.mag)
        assert_quantity_allclose(tbl[0]['flux_scaling'], 3631 * u.Jy)
        assert_quantity_allclose(tbl[0]['min'], 0.041017 * data_unit, atol=1e-5 * data_unit)
        assert_quantity_allclose(tbl[0]['max'], 138.923752 * data_unit, rtol=1e-5)
        assert_quantity_allclose(tbl[0]['mean'], 4.391718 * data_unit)
        assert_quantity_allclose(tbl[0]['median'], 0.482972 * data_unit, rtol=1e-5)
        assert_quantity_allclose(tbl[0]['mode'], -7.33452046 * data_unit)
        assert_quantity_allclose(tbl[0]['std'], 15.618626488031158 * data_unit)
        assert_quantity_allclose(tbl[0]['mad_std'], 0.47892631 * data_unit)
        assert_quantity_allclose(tbl[0]['var'], 243.9415 * (data_unit * data_unit))
        assert_quantity_allclose(tbl[0]['biweight_location'], 0.40664572 * data_unit)
        assert_quantity_allclose(tbl[0]['biweight_midvariance'], 0.27319583 * (data_unit * data_unit))  # noqa
        assert_quantity_allclose(tbl[0]['fwhm'], 2.8691718 * u.pix)
        assert_quantity_allclose(tbl[0]['semimajor_sigma'], 1.22308648 * u.pix)
        assert_quantity_allclose(tbl[0]['semiminor_sigma'], 1.21374578 * u.pix)
        assert_quantity_allclose(tbl[0]['orientation'], -27.220559 * u.deg)
        assert_quantity_allclose(tbl[0]['eccentricity'], 0.12335181)

        # --- Back to parser testing below. ---

        # Request specific extension (name + ver, but ver is not used), use given label
        imviz_helper.load_data(self.jwst_asdf_url_1, cache=True, ext=('DQ', 42),
                               data_label='jw01072001001_01101_00001_nrcb1_cal',
                               show_in_viewer=False)
        data = imviz_helper.app.data_collection[1]
        comp = data.get_component('DQ')
        assert data.label == 'jw01072001001_01101_00001_nrcb1_cal[DQ]'
        assert data.meta['aperture']['name'] == 'NRCB5_FULL'
        assert comp.units == ''

        # Pass in HDUList directly + ext (name only), use given label
        filename = download_file(self.jwst_asdf_url_1, cache=True)
        with fits.open(filename) as pf:
            imviz_helper.load_data(pf, ext='SCI',
                                   data_label='jw01072001001_01101_00001_nrcb1_cal',
                                   show_in_viewer=False)
            data = imviz_helper.app.data_collection[2]
            comp = data.get_component('DATA')  # SCI = DATA
            assert data.label == 'jw01072001001_01101_00001_nrcb1_cal[DATA]'
            assert isinstance(data.coords, GWCS)
            assert comp.units == 'MJy/sr'

            # Test duplicate label functionality
            imviz_helper.app.data_collection.clear()
            imviz_helper.load_data(pf, ext='SCI', data_label='TEST', show_in_viewer=False)
            data = imviz_helper.app.data_collection[0]
            assert data.label.endswith('[DATA]')

            imviz_helper.load_data(pf, ext='SCI', data_label='TEST', show_in_viewer=False)
            data = imviz_helper.app.data_collection[1]
            assert data.label.endswith('[DATA] (1)')

            # Load all extensions
            imviz_helper.app.data_collection.clear()
            imviz_helper.load_data(pf, ext='*', show_in_viewer=False)
            data = imviz_helper.app.data_collection
            assert len(data.labels) == 7
            assert data.labels[0].endswith('[DATA]')
            assert data.labels[1].endswith('[ERR]')
            assert data.labels[2].endswith('[DQ]')
            assert data.labels[3].endswith('[AREA]')
            assert data.labels[4].endswith('[VAR_POISSON]')
            assert data.labels[5].endswith('[VAR_RNOISE]')
            assert data.labels[6].endswith('[VAR_FLAT]')

        # Invalid ASDF attribute (extension)
        with pytest.raises(KeyError, match='does_not_exist'):
            parse_data(imviz_helper.app, filename, ext='DOES_NOT_EXIST', data_label='foo')

    @pytest.mark.remote_data
    def test_parse_jwst_niriss_grism(self, imviz_helper):
        """No valid image GWCS for Imviz, will fall back to loading without WCS."""
        imviz_helper.load_data(self.jwst_asdf_url_2, cache=True, show_in_viewer=False)
        data = imviz_helper.app.data_collection[0]
        comp = data.get_component('DATA')
        expected_label = os.path.splitext(os.path.basename(self.jwst_asdf_url_2))[0] + '[DATA]'
        assert data.label == expected_label
        assert data.shape == (2048, 2048)
        assert data.coords is None
        assert comp.units == 'DN/s'
        assert comp.data.shape == (2048, 2048)

    @pytest.mark.remote_data
    def test_parse_hst_drz(self, imviz_helper):
        url = 'https://mast.stsci.edu/api/v0.1/Download/file?bundle_name=MAST_2021-04-21T1828.sh&uri=mast:HST/product/jclj01010_drz.fits'  # noqa: E501
        filename = download_file(url, cache=True)

        # Default behavior: Load first image
        imviz_helper.load_data(url, cache=True)
        data = imviz_helper.app.data_collection[0]
        comp = data.get_component('SCI,1')
        expected_label = os.path.splitext(os.path.basename(url))[0] + '[SCI,1]'
        assert data.label == expected_label
        assert data.shape == (4299, 4219)
        assert_allclose(data.meta['PHOTFLAM'], 7.8711728E-20)
        assert isinstance(data.coords, WCS)
        assert comp.units == 'electron/s'
        assert comp.data.shape == (4299, 4219)

        # --- Since download is expensive, we attach FITS WCS-specific tests here. ---

        # Test simple aperture photometry plugin.
        imviz_helper.plugins['Subset Tools'].import_region(
            EllipsePixelRegion(center=PixCoord(x=1488.5, y=2576), width=47, height=70))  # Galaxy
        phot_plugin = imviz_helper.app.get_tray_item_from_name('imviz-aper-phot-simple')
        phot_plugin.data_selected = 'contents[SCI,1]'
        phot_plugin.aperture_selected = 'Subset 1'
        assert phot_plugin.aperture.selected_validity.get('is_aperture')
        phot_plugin.background_value = 0.0014  # Manual entry: Median on whole array
        assert_allclose(phot_plugin.pixel_area, 0.0025)  # Not used but still auto-populated
        phot_plugin.vue_do_aper_phot()
        tbl = imviz_helper.plugins['Aperture Photometry'].export_table()
        assert_quantity_allclose(tbl[0]['xcenter'], 1488.5 * u.pix, atol=2 * u.pix)
        assert_quantity_allclose(tbl[0]['ycenter'], 2576 * u.pix, atol=2 * u.pix)
        sky = tbl[0]['sky_center']
        assert_allclose(sky.ra.deg, 3.684062989070131, rtol=1e-3)
        assert_allclose(sky.dec.deg, 10.802045612042956, rtol=1e-3)
        data_unit = u.electron / u.s
        assert_quantity_allclose(tbl[0]['background'], 0.0014 * data_unit)
        assert_quantity_allclose(tbl[0]['sum'], 112.712406 * data_unit, rtol=1e-3)
        assert_quantity_allclose(tbl[0]['sum_aper_area'], 2583.959958 * PIX2, rtol=1e-3)
        assert_array_equal(tbl[0]['pixarea_tot'], None)
        assert_array_equal(tbl[0]['aperture_sum_counts'], None)
        assert_array_equal(tbl[0]['aperture_sum_counts_err'], None)
        assert_array_equal(tbl[0]['counts_fac'], None)
        assert_array_equal(tbl[0]['aperture_sum_mag'], None)
        assert_array_equal(tbl[0]['flux_scaling'], None)
        assert_quantity_allclose(tbl[0]['min'], -0.02422 * data_unit, rtol=1e-3)
        assert_quantity_allclose(tbl[0]['max'], 1.577081 * data_unit, rtol=1e-3)
        assert_quantity_allclose(tbl[0]['mean'], 0.043684 * data_unit, rtol=1e-3)
        assert_quantity_allclose(tbl[0]['median'], 0.02129 * data_unit, rtol=1e-3)
        assert_quantity_allclose(tbl[0]['mode'], -0.023497 * data_unit, rtol=1e-3)
        assert_quantity_allclose(tbl[0]['std'], 0.09908 * data_unit, rtol=1e-3)
        assert_quantity_allclose(tbl[0]['mad_std'], 0.023219 * data_unit, rtol=1e-3)
        assert_quantity_allclose(tbl[0]['var'], 0.009817 * (data_unit * data_unit), rtol=1e-3)
        assert_quantity_allclose(tbl[0]['biweight_location'], 0.020978 * data_unit, rtol=1e-3)
        assert_quantity_allclose(tbl[0]['biweight_midvariance'], 0.000606 * (data_unit * data_unit), rtol=1e-3)  # noqa
        assert_quantity_allclose(tbl[0]['fwhm'], 21.230628 * u.pix, rtol=1e-3)
        assert_quantity_allclose(tbl[0]['semimajor_sigma'], 9.712415 * u.pix, rtol=1e-3)
        assert_quantity_allclose(tbl[0]['semiminor_sigma'], 8.260686 * u.pix, rtol=1e-3)
        assert_quantity_allclose(tbl[0]['orientation'], 80.054019 * u.deg, rtol=1e-3)
        assert_quantity_allclose(tbl[0]['eccentricity'], 0.525929, rtol=1e-3)

        # Request specific extension (name only), use given label
        imviz_helper.load_data(filename, ext='CTX', data_label='jclj01010_drz',
                               show_in_viewer=False)
        data = imviz_helper.app.data_collection[1]
        comp = data.get_component('CTX,1')
        assert data.label == 'jclj01010_drz[CTX,1]'
        assert data.meta['EXTNAME'] == 'CTX'
        assert comp.units == ''  # BUNIT is not set

        # Request specific extension (name + ver), use given label
        imviz_helper.load_data(filename, ext=('WHT', 1), data_label='jclj01010_drz',
                               show_in_viewer=False)
        data = imviz_helper.app.data_collection[2]
        comp = data.get_component('WHT,1')
        assert data.label == 'jclj01010_drz[WHT,1]'
        assert data.meta['EXTNAME'] == 'WHT'
        assert comp.units == ''  # BUNIT is not set

        # Pass in file obj directly
        with fits.open(filename) as pf:
            # Default behavior: Load first image
            imviz_helper.load_data(pf, show_in_viewer=False)
            data = imviz_helper.app.data_collection[3]
            assert data.label.startswith('Unknown HDU object') and data.label.endswith('[SCI,1]')
            assert_allclose(data.meta['PHOTFLAM'], 7.8711728E-20)
            assert 'SCI,1' in data.components

            # Request specific extension (name only), use given label
            imviz_helper.load_data(pf, ext='CTX', show_in_viewer=False)
            data = imviz_helper.app.data_collection[4]
            assert data.label.startswith('Unknown HDU object') and data.label.endswith('[CTX,1]')
            assert data.meta['EXTNAME'] == 'CTX'
            assert 'CTX,1' in data.components

            # Pass in HDU directly, use given label
            imviz_helper.load_data(pf[2], data_label='foo', show_in_viewer=False)
            data = imviz_helper.app.data_collection[5]
            assert data.label == 'foo[WHT,1]'
            assert data.meta['EXTNAME'] == 'WHT'
            assert 'WHT,1' in data.components

            # Load all extensions
            imviz_helper.app.data_collection.clear()
            imviz_helper.load_data(filename, ext='*', show_in_viewer=False)
            data = imviz_helper.app.data_collection
            assert len(data.labels) == 3
            assert data.labels[0].endswith('[SCI,1]')
            assert data.labels[1].endswith('[WHT,1]')
            assert data.labels[2].endswith('[CTX,1]')

        # Cannot load non-image extension
        with pytest.raises(ValueError, match='Imviz cannot load this HDU'):
            parse_data(imviz_helper.app, filename, ext='HDRTAB')

        # Invalid FITS extension
        with pytest.raises(KeyError, match='not found'):
            parse_data(imviz_helper.app, filename, ext='DOES_NOT_EXIST', data_label='foo')


def test_load_valid_not_valid(imviz_helper):
    # Load something valid.
    arr = np.ones((5, 5))
    imviz_helper.load_data(arr, data_label='valid', show_in_viewer=False)

    # Load something invalid.
    with pytest.raises(ValueError, match='Imviz cannot load this array with ndim=1'):
        imviz_helper.load_data(np.zeros(2), show_in_viewer=False)

    # Make sure valid data is still there.
    assert (len(imviz_helper.app.data_collection) == 1
            and imviz_helper.app.data_collection.labels == ['valid'])
    assert_allclose(imviz_helper.app.data_collection[0].get_component('DATA').data, 1)


@pytest.mark.skipif(not HAS_ROMAN_DATAMODELS, reason="roman_datamodels is not installed")
def test_roman_parser(imviz_helper, roman_imagemodel):
    imviz_helper.load_data(roman_imagemodel, data_label='roman_wfi_image_model', ext='data')
    assert len(imviz_helper.app.data_collection) == 1



import pytest

import numpy as np
from astropy import units as u
from astropy.wcs import WCS
from specutils import Spectrum1D
from jdaviz import Application, Specviz
from jdaviz.configs.default.plugins.gaussian_smooth.gaussian_smooth import GaussianSmooth
from jdaviz.core.unit_conversion_utils import (flux_conversion_general,
                                               viewer_flux_conversion_equivalencies)


# This applies to all viz but testing with Imviz should be enough.
def test_viewer_calling_app(imviz_helper):
    viewer = imviz_helper.default_viewer._obj
    assert viewer.session.jdaviz_app is imviz_helper.app


def test_get_tray_item_from_name():
    app = Application(configuration='default')
    plg = app.get_tray_item_from_name('g-gaussian-smooth')
    assert isinstance(plg, GaussianSmooth)

    with pytest.raises(KeyError, match='not found in app'):
        app.get_tray_item_from_name('imviz-compass')


def test_nonstandard_specviz_viewer_name(spectrum1d):
    config = {'settings': {'configuration': 'nonstandard',
                           'data': {'parser': 'specviz-spectrum1d-parser'},
                           'visible': {'menu_bar': False,
                                       'toolbar': True,
                                       'tray': True,
                                       'tab_headers': False},
                           'context': {'notebook': {'max_height': '750px'}}},
              'toolbar': ['g-data-tools', 'g-subset-tools'],
              'tray': ['g-metadata-viewer',
                       'g-plot-options',
                       'g-subset-tools',
                       'g-gaussian-smooth',
                       'g-model-fitting',
                       'g-unit-conversion',
                       'g-line-list',
                       'specviz-line-analysis',
                       'export'],
              'viewer_area': [{'container': 'col',
                               'children': [{'container': 'row',
                                             'viewers': [{'name': 'H',
                                                          'plot': 'specviz-profile-viewer',
                                                          'reference': 'h'},
                                                         {'name': 'K',
                                                          'plot': 'specviz-profile-viewer',
                                                          'reference': 'k'}]}]}]}

    class Customviz(Specviz):
        _default_configuration = config
        _default_spectrum_viewer_reference_name = 'h'

    viz = Customviz()
    assert viz.app.get_viewer_reference_names() == ['h', 'k']

    viz.load_data(spectrum1d, data_label='example label')
    with pytest.raises(ValueError):
        viz.get_data("non-existent label")


def test_duplicate_data_labels(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d, data_label="test")
    specviz_helper.load_data(spectrum1d, data_label="test")
    dc = specviz_helper.app.data_collection
    assert dc[0].label == "test"
    assert dc[1].label == "test (1)"
    specviz_helper.load_data(spectrum1d, data_label="test_1")
    specviz_helper.load_data(spectrum1d, data_label="test")
    assert dc[2].label == "test_1"
    assert dc[3].label == "test (2)"


def test_duplicate_data_labels_with_brackets(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d, data_label="test[test]")
    specviz_helper.load_data(spectrum1d, data_label="test[test]")
    dc = specviz_helper.app.data_collection
    assert len(dc) == 2
    assert dc[0].label == "test[test]"
    assert dc[1].label == "test[test] (1)"


def test_return_data_label_is_none(specviz_helper):
    data_label = specviz_helper.app.return_data_label(None)
    assert data_label == "Unknown"


def test_return_data_label_is_image(specviz_helper):
    data_label = specviz_helper.app.return_data_label("data/path/test.jpg")
    assert data_label == "test[jpg]"


def test_hdulist_with_filename(cubeviz_helper, image_cube_hdu_obj):
    image_cube_hdu_obj.file_name = "test"
    data_label = cubeviz_helper.app.return_data_label(image_cube_hdu_obj)
    assert data_label == "test[HDU object]"


def test_file_path_not_image(imviz_helper, tmp_path):
    path = tmp_path / "myimage.fits"
    path.touch()
    data_label = imviz_helper.app.return_data_label(str(path))
    assert data_label == "myimage"


def test_unique_name_variations(specviz_helper, spectrum1d):
    data_label = specviz_helper.app.return_unique_name(None)
    assert data_label == "Unknown"

    specviz_helper.load_data(spectrum1d, data_label="test[flux]")
    data_label = specviz_helper.app.return_data_label("test[flux]", ext="flux")
    assert data_label == "test[flux][flux]"

    data_label = specviz_helper.app.return_data_label("test", ext="flux")
    assert data_label == "test[flux] (1)"


def test_substring_in_label(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d, data_label="M31")
    specviz_helper.load_data(spectrum1d, data_label="M32")
    data_label = specviz_helper.app.return_data_label("M")
    assert data_label == "M"


@pytest.mark.parametrize('data_label', ('111111', 'aaaaa', '///(#$@)',
                                        'two  spaces  repeating',
                                        'word42word42word  two  spaces'))
def test_edge_cases(specviz_helper, spectrum1d, data_label):
    dc = specviz_helper.app.data_collection

    specviz_helper.load_data(spectrum1d, data_label=data_label)
    specviz_helper.load_data(spectrum1d, data_label=data_label)
    assert dc[1].label == f"{data_label} (1)"

    specviz_helper.load_data(spectrum1d, data_label=data_label)
    assert dc[2].label == f"{data_label} (2)"


def test_case_that_used_to_break_return_label(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d, data_label="this used to break (1)")
    specviz_helper.load_data(spectrum1d, data_label="this used to break")
    dc = specviz_helper.app.data_collection
    assert dc[0].label == "this used to break (1)"
    assert dc[1].label == "this used to break (2)"


def test_viewer_renaming_specviz(specviz_helper):
    viewer_names = [
        'spectrum-viewer',
        'second-viewer-name',
        'third-viewer-name'
    ]

    for i in range(len(viewer_names) - 1):
        specviz_helper.app._update_viewer_reference_name(
            old_reference=viewer_names[i],
            new_reference=viewer_names[i + 1]
        )
        assert specviz_helper.app.get_viewer(viewer_names[i+1]) is not None


def test_viewer_renaming_imviz(imviz_helper):
    with pytest.raises(ValueError, match="'imviz-0' cannot be changed"):
        imviz_helper.app._update_viewer_reference_name(
            old_reference='imviz-0',
            new_reference='this-is-forbidden'
        )

    with pytest.raises(ValueError, match="does not exist"):
        imviz_helper.app._update_viewer_reference_name(
            old_reference='non-existent',
            new_reference='this-is-forbidden'
        )


def test_data_associations(imviz_helper):
    shape = (10, 10)

    data_parent = np.ones(shape, dtype=float)
    data_child = np.zeros(shape, dtype=int)

    imviz_helper.load_data(data_parent, data_label='parent_data')
    imviz_helper.load_data(data_child, data_label='child_data', parent='parent_data')

    assert imviz_helper.app._get_assoc_data_children('parent_data') == ['child_data']
    assert imviz_helper.app._get_assoc_data_parent('child_data') == 'parent_data'

    with pytest.raises(NotImplementedError):
        # we don't (yet) allow children of children:
        imviz_helper.load_data(data_child, data_label='grandchild_data', parent='child_data')

    with pytest.raises(ValueError):
        # ensure the parent actually exists:
        imviz_helper.load_data(data_child, data_label='child_data', parent='absent parent')


def test_to_unit(cubeviz_helper):
    # custom cube to have Surface Brightness units
    wcs_dict = {"CTYPE1": "WAVE-LOG", "CTYPE2": "DEC--TAN", "CTYPE3": "RA---TAN",
                "CRVAL1": 4.622e-7, "CRVAL2": 27, "CRVAL3": 205,
                "CDELT1": 8e-11, "CDELT2": 0.0001, "CDELT3": -0.0001,
                "CRPIX1": 0, "CRPIX2": 0, "CRPIX3": 0, "PIXAR_SR": 8e-11}
    w = WCS(wcs_dict)
    flux = np.zeros((30, 20, 3001), dtype=np.float32)
    flux[5:15, 1:11, :] = 1
    cube = Spectrum1D(flux=flux * (u.MJy / u.sr), wcs=w, meta=wcs_dict)
    cubeviz_helper.load_data(cube, data_label="test")

    # this can be removed once spectra pass through spectral extraction
    extract_plg = cubeviz_helper.plugins['Spectral Extraction']

    extract_plg.aperture = extract_plg.aperture.choices[-1]
    extract_plg.aperture_method.selected = 'Exact'
    extract_plg.wavelength_dependent = True
    extract_plg.function = 'Sum'
    # set so pixel scale factor != 1
    extract_plg.reference_spectral_value = 0.000001

    extract_plg.extract()

    data = cubeviz_helper.app.data_collection[-1].data
    values = [1]
    original_units = u.MJy / u.sr
    target_units = u.MJy

    spec = data.get_object(cls=Spectrum1D)
    viewer_equivs = viewer_flux_conversion_equivalencies(values, spec)
    value = flux_conversion_general(values, original_units,
                                    target_units, viewer_equivs,
                                    with_unit=False)

    # will be a uniform array since not wavelength dependent
    # so test first value in array
    assert np.allclose(value[0], 8e-11)

    # Change from Fnu to Flam (with values shape matching spectral axis)

    values = np.ones(3001)
    original_units = u.MJy
    target_units = u.erg / u.cm**2 / u.s / u.AA

    viewer_equivs = viewer_flux_conversion_equivalencies(values, spec)
    new_values = flux_conversion_general(values, original_units,
                                         target_units, viewer_equivs,
                                         with_unit=False)

    assert np.allclose(new_values,
                       (values * original_units)
                       .to_value(target_units,
                                 equivalencies=u.spectral_density(cube.spectral_axis)))

    # Change from Fnu to Flam (with a shape (2,) array of values indicating we
    # are probably converting the limits)

    values = [1, 2]
    original_units = u.MJy
    target_units = u.erg / u.cm**2 / u.s / u.AA

    viewer_equivs = viewer_flux_conversion_equivalencies(values, spec)
    new_values = flux_conversion_general(values, original_units,
                                         target_units, viewer_equivs,
                                         with_unit=False)

    # In this case we do a regular spectral density conversion, but using the
    # first value in the spectral axis for the equivalency
    assert np.allclose(new_values,
                       ([1, 2] * original_units)
                       .to_value(target_units,
                                 equivalencies=u.spectral_density(cube.spectral_axis[0])))


def test_all_plugins_have_description(cubeviz_helper, specviz_helper,
                                      mosviz_helper, imviz_helper,
                                      rampviz_helper, specviz2d_helper):
    """
    Test that all plugins for all configs have a plugin_description
    attribute, which controls what is displayed under the plugin title in the
    tray. Doesn't test what they are, just that they are not empty.
    """

    config_helpers = [cubeviz_helper, specviz_helper, mosviz_helper,
                      imviz_helper, rampviz_helper, specviz2d_helper]

    for config_helper in config_helpers:
        for item in config_helper.plugins:
            assert config_helper.plugins[item]._obj.plugin_description != ''



# adapted from https://github.com/spacetelescope/jwst_novt/blob/main/jwst_novt/footprints.py

import numpy as np
import regions
from astropy import coordinates

try:
    import pysiaf
except ImportError:
    _has_pysiaf = False
else:
    _has_pysiaf = True

__all__ = [
    "_has_pysiaf",
    "_instruments",
    "_full_apertures",
    "_all_apertures",
    "instrument_footprint"
]

_instruments = {
    'JWST': {
        'NIRSpec': 'NIRSpec',
        'NIRCam:short': 'NIRCam',
        'NIRCam:long': 'NIRCam',
        'NIRISS': 'NIRISS',
        'MIRI': 'MIRI',
        'FGS': 'FGS',
    },
    'Roman': {
        'WFI': 'Roman',
        'WFI+CGI': 'Roman',
    }
}

_full_apertures = {'NIRSpec': 'NRS_FULL_MSA',
                   'NIRCam:short': 'NRCALL_FULL',
                   'NIRCam:long': 'NRCALL_FULL',
                   'NIRISS': 'NIS_AMIFULL',
                   'MIRI': 'MIRIM_FULL',
                   'FGS': 'FGS1_FULL',
                   }

_all_apertures = {'NIRSpec': ["NRS_FULL_MSA1",
                              "NRS_FULL_MSA2",
                              "NRS_FULL_MSA3",
                              "NRS_FULL_MSA4",
                              "NRS_FULL_IFU",
                              "NRS_S200A1_SLIT",
                              "NRS_S200A2_SLIT",
                              "NRS_S400A1_SLIT",
                              "NRS_S1600A1_SLIT",
                              "NRS_S200B1_SLIT"],
                  'NIRCam:short': ["NRCA1_FULL",
                                   "NRCA2_FULL",
                                   "NRCA3_FULL",
                                   "NRCA4_FULL",
                                   "NRCB1_FULL",
                                   "NRCB2_FULL",
                                   "NRCB3_FULL",
                                   "NRCB4_FULL"],
                  'NIRCam:long': ["NRCA5_FULL", "NRCB5_FULL"],
                  'NIRISS': ['NIS_AMIFULL'],
                  'MIRI': ['MIRIM_FULL'],
                  'FGS': ['FGS1_FULL', 'FGS2_FULL'],
                  'WFI': [f'WFI{i+1:02d}_FULL' for i in range(18)],
                  'WFI+CGI': [f'WFI{i+1:02d}_FULL' for i in range(18)] + ['CGI_CEN'],
                  }


def instrument_footprint(
        observatory, instrument, ra, dec, pa,
        v2_offset=0.0, v3_offset=0.0, apertures=None
):
    """
    Create footprint regions in sky coordinates from a jwst instrument.

    Parameters
    ----------
    observatory : string
        Observatory, one of {"JWST", "Roman"}.
    instrument : string
        Instrument. For JWST, must be one of
        {"NIRSpec", "NIRCam:short", "NIRCam:long", "NIRISS", "MIRI", "FGS"};
        for Roman, must be one of {"WFI", "CGI"}.
    ra : float
        RA of NIRCam center, in degrees.
    dec : float
        Dec of NIRCam center, in degrees.
    pa : float
        Position angle, in degrees measured from North
        to central vertical axis in North to East direction.
    v2_offset : float, optional
        Additional V2 offset in telescope coordinates to apply to instrument
        center, as from a dither pattern.
    v3_offset : float, optional
        Additional V3 offset in telescope coordinates to apply to instrument
        center, as from a dither pattern.
    apertures : list of str, optional
        If set, only the specified apertures are returned.

    Returns
    -------
    footprint : regions.Regions
        Footprint regions as Polygon regions in sky coordinates.
    """
    if not _has_pysiaf:
        raise ImportError('jwst_footprint requires pysiaf to be installed')

    if observatory not in _instruments:
        raise ValueError(f"observatory must be one of {', '.join(_instruments.keys())}")

    if instrument not in _instruments[observatory]:  # pragma: no cover
        raise ValueError(f"instrument must be one of {', '.join(_instruments[observatory].keys())}")

    siaf_interface = pysiaf.Siaf(_instruments.get(observatory).get(instrument))

    # use different definitions of the center for JWST and Roman:
    if observatory == "Roman":
        if 'CGI' not in instrument:
            center = siaf_interface.apertures['WFI_CEN']
        else:
            center = siaf_interface.apertures['CGI_CEN']
        v2 = center.V2Ref - v2_offset
        v3 = center.V3Ref + v3_offset
        pa_offset = center.V3IdlYAngle
    else:
        # Get center and PA offset from full aperture
        full = siaf_interface.apertures[_full_apertures.get(instrument)]
        corners = full.corners("tel", rederive=False)
        v2 = np.mean(corners[0]) - v2_offset
        v3 = np.mean(corners[1]) + v3_offset
        pa_offset = full.V3IdlYAngle

    # Attitude matrix for sky coordinates
    attmat = pysiaf.utils.rotations.attitude(v2, v3, ra, dec, pa - pa_offset)

    if apertures is None:
        apertures = _all_apertures.get(instrument)

    # Aperture regions
    ap_regions = []
    for aperture_name in apertures:
        aperture = siaf_interface.apertures[aperture_name]
        aperture.set_attitude_matrix(attmat)
        poly_points = aperture.closed_polygon_points(
            "sky", rederive='CGI' not in instrument
        )

        sky_coord = coordinates.SkyCoord(*poly_points, unit="deg")
        reg = regions.PolygonSkyRegion(sky_coord)
        ap_regions.append(reg)

    return regions.Regions(ap_regions)



from collections.abc import Iterable
import csv
import os
from pathlib import Path
import warnings

from astropy import units as u
from astropy.io import fits
from astropy.io.registry import IORegistryError
from astropy.wcs import WCS
from glue.core.data import Data
from glue.core.link_helpers import LinkSameWithUnits
from specutils import Spectrum1D, SpectrumList, SpectrumCollection
from specutils.io.default_loaders.jwst_reader import identify_jwst_s2d_multi_fits

from jdaviz.configs.imviz.plugins.parsers import get_image_data_iterator
from jdaviz.core.registries import data_parser_registry
from jdaviz.core.events import SnackbarMessage
from jdaviz.utils import standardize_metadata, PRIHDR_KEY, download_uri_to_path

__all__ = ['mos_spec1d_parser', 'mos_spec2d_parser', 'mos_image_parser']

FALLBACK_NAME = "Unspecified"
EXPECTED_FILES = {"niriss": ['1D Spectra C', '1D Spectra R',
                             '2D Spectra C', '2D Spectra R',
                             'Direct Image'],
                  "nircam": ['1D Spectra C', '1D Spectra R',
                             '2D Spectra C', '2D Spectra R',
                             'Direct Image'],
                  "nirspec": ['1D Spectra', '2D Spectra']}


def _add_to_table(app, data, comp_label, table_viewer_reference_name='table-viewer'):
    """
    Creates a mos table instance in the application data collection is none
    currently exists.

    Parameters
    ----------
    app : `~jdaviz.app.Application`
        The Jdaviz application instance.
    data : array-list
        The set of data to added as a table (i.g. column) component.
    comp_label : str
        The label used to describe the data. Also used as the column header.
    """
    # Add data to the mos viz table object
    if data is None or len(data) == 0:
        return

    if 'MOS Table' not in app.data_collection:
        table_data = Data(label='MOS Table')
        app.add_data(table_data, notify_done=False)

        mos_table = app.data_collection['MOS Table']
        mos_table.add_component(data, comp_label)

        viewer = app.get_viewer(table_viewer_reference_name)
        viewer.add_data(table_data)
    else:
        mos_table = app.data_collection['MOS Table']
        if comp_label not in mos_table.component_ids():
            mos_table.add_component(data, comp_label)


def _check_is_file(path):
    return isinstance(path, str) and Path(path).is_file()


def _warn_if_not_found(app, files_by_labels):
    """
    Take a list of labels and associated file lists/strings and send a
    snackbar message if any are empty.
    """
    found = []
    not_found = []
    for key in files_by_labels:
        if files_by_labels[key]:
            found.append(key)
        else:
            not_found.append(key)

    if len(found) == 0:
        raise ValueError("No valid files found in specified directory")
    if len(not_found) > 0:
        warn_msg = f"Some files not found: {', '.join(not_found)}"
        warn_msg = SnackbarMessage(warn_msg, color="warning", sender=app)
        app.hub.broadcast(warn_msg)

    return found


def _fields_from_ecsv(fname, fields, delimiter=","):
    """
    Save specified field(s) from an ecsv file as a row-by-row list of
    lists.
    """
    parsed_fields = []
    with open(fname, "r") as f:
        reader = csv.DictReader(filter(lambda r: r[0] != "#", f),
                                delimiter=delimiter)
        for row in reader:
            temp_list = []
            for field in fields:
                temp_list.append(row[field])
            parsed_fields.append(temp_list)
    return parsed_fields


@data_parser_registry("mosviz-link-data")
def link_data_in_table(app, data_obj=None):
    """
    Batch links data in the mosviz table viewer.

    Parameters
    ----------
    app : `~jdaviz.app.Application`
        The application-level object used to reference the viewers.
    data_obj : None
        Passed in in order to use the data_parser_registry, otherwise
        not used.
    """
    mos_data = app.session.data_collection['MOS Table']
    wc_spec_ids = []

    # Optimize linking speed through a) delaying link manager updates with a
    # context manager, b) handling intra-row linkage of 1D and 2D spectra in a
    # loop, and c) handling inter-row linkage after that in one fell swoop.
    with app.data_collection.delay_link_manager_update():

        spectra_1d = mos_data.get_component('1D Spectra').data
        spectra_2d = mos_data.get_component('2D Spectra').data

        # Link each 1D spectrum with its corresponding 2D spectrum
        for index in range(len(spectra_1d)):

            spec_1d = spectra_1d[index]
            spec_2d = spectra_2d[index]

            wc_spec_1d = app.session.data_collection[spec_1d].world_component_ids
            wc_spec_2d = app.session.data_collection[spec_2d].world_component_ids

            wc_spec_ids.append(LinkSameWithUnits(wc_spec_1d[0], wc_spec_2d[1]))

    app.session.data_collection.add_link(wc_spec_ids)


@data_parser_registry("mosviz-nirspec-directory-parser")
def mos_nirspec_directory_parser(app, data_obj, data_labels=None):

    spectra_1d = []
    spectra_2d = []

    # Load spectra
    level3_path = Path(data_obj)
    for p in sorted(level3_path.glob('*.fits*')):
        file_path = str(p)
        if 'x1d' in file_path or 'c1d' in file_path:
            spectra_1d.append(file_path)
        elif 's2d' in file_path:
            spectra_2d.append(file_path)

    n_specs = mos_spec1d_parser(app, spectra_1d)
    mos_spec2d_parser(app, spectra_2d)

    # Load images, if present
    image_path = None

    # Potential names of subdirectories where images are stored
    for image_dir_name in ("cutouts", "mosviz_cutouts", "images"):
        cur_path = level3_path / image_dir_name
        if cur_path.is_dir():
            image_path = cur_path
            break
    if image_path is not None:
        images = sorted(image_path.glob('*.fits*'))
        n_images = len(images)

        # The amount of images needs to be equal to the amount of rows
        # of the other columns in the table
        if n_images == 1:
            app._jdaviz_helper._shared_image = True
            app.get_viewer(app._jdaviz_helper._default_table_viewer_reference_name)._shared_image = True  # noqa: E501
            if n_specs > 1:
                kwargs = {'share_image': n_specs}
            else:
                kwargs = {}
            mos_image_parser(app, str(images[0]), **kwargs)
        elif n_images == n_specs:
            mos_image_parser(app, list(map(str, images)))
        else:
            app.hub.broadcast(SnackbarMessage(
                "The number of images in this directory does not match the "
                "number of spectra 1d and 2d files, please make the "
                "amounts equal or load images separately.", color='warning', sender=app))

    mos_meta_parser(app)


@data_parser_registry("mosviz-spec1d-parser")
def mos_spec1d_parser(app, data_obj, data_labels=None,
                      table_viewer_reference_name='table-viewer'):
    """
    Attempts to parse a 1D spectrum object.

    Parameters
    ----------
    app : `~jdaviz.app.Application`
        The application-level object used to reference the viewers.
    data_obj : str or list or spectrum-like
        File path, list, or spectrum-like object to be read as a new row in
        the mosviz table.
    data_labels : str, optional
        The label applied to the glue data component.

    Returns
    -------
    n_specs : int
        Number of data objects loaded.

    """
    if isinstance(data_labels, str):
        data_labels = [data_labels]

    # Coerce into list if needed
    if not isinstance(data_obj, (list, tuple, SpectrumCollection)):
        data_obj = [data_obj]

    # If the file has multiple objects in it, the Spectrum1D read machinery
    # will fail to find a reader for it, and we fall back on SpectrumList
    try:
        data_obj = [Spectrum1D.read(x) if _check_is_file(x) else x for x in data_obj]
    except IORegistryError:
        if len(data_obj) == 1:
            if _check_is_file(data_obj[0]):
                data_obj = SpectrumList.read(data_obj[0])

    if data_labels is None:
        data_labels = [f"1D Spectrum {i}" for i in range(len(data_obj))]
    elif len(data_obj) != len(data_labels):
        data_labels = [f"{data_labels[0]} {i}" for i in range(len(data_obj))]

    # Handle the case where the 1d spectrum is a collection of spectra

    with app.data_collection.delay_link_manager_update():

        for i, (cur_data, cur_label) in enumerate(zip(data_obj, data_labels)):
            # Make metadata layout conform with other viz.
            cur_data.meta = standardize_metadata(cur_data.meta)
            cur_data.meta['mosviz_row'] = i

            app.add_data(cur_data, cur_label, notify_done=False)

        _add_to_table(app, data_labels, '1D Spectra',
                      table_viewer_reference_name=table_viewer_reference_name)

    return len(data_obj)


@data_parser_registry("mosviz-spec2d-parser")
def mos_spec2d_parser(app, data_obj, data_labels=None, add_to_table=True,
                      show_in_viewer=False, ext=1, transpose=False,
                      cache=None, local_path=None, timeout=None):
    """
    Attempts to parse a 2D spectrum object.

    Notes
    -----
    Default arguments assume that the data is in the second HDU of the FITS file unless
    otherwise specified with the ``ext`` parameter.

    Parameters
    ----------
    app : `~jdaviz.app.Application`
        The application-level object used to reference the viewers.
    data_obj : str or list or spectrum-like
        File path, list, or spectrum-like object to be read as a new row in
        the mosviz table.
    data_labels : str, optional
        The label applied to the glue data component.
    ext : int, optional
        The extension in the FITS file that contains the data to be loaded.
    transpose : bool, optional
        Flag to transpose the data array before loading.
    cache : None, bool, or str
        Cache the downloaded file if the data are retrieved by a query
        to a URL or URI.
    local_path : str, optional
        Cache remote files to this path. This is only used if data is
        requested from `astroquery.mast`.
    timeout : float, optional
        If downloading from a remote URI, set the timeout limit for
        remote requests in seconds (passed to
        `~astropy.utils.data.download_file` or
        `~astroquery.mast.Conf.timeout`).

    Returns
    -------
    n_specs : int
        Number of data objects loaded.

    """
    spectrum_2d_viewer_reference_name = (
        app._jdaviz_helper._default_spectrum_2d_viewer_reference_name
    )
    table_viewer_reference_name = (
        app._jdaviz_helper._default_table_viewer_reference_name
    )

    # Note: This is also used by Specviz2D
    def _parse_as_spectrum1d(hdulist, ext, transpose):
        # Parse as a FITS file and assume the WCS is correct
        data = hdulist[ext].data
        header = hdulist[ext].header
        metadata = standardize_metadata(header)
        metadata[PRIHDR_KEY] = standardize_metadata(hdulist[0].header)
        wcs = WCS(header, hdulist)
        if transpose:
            data = data.T
            wcs = wcs.swapaxes(0, 1)

        try:
            data_unit = u.Unit(header['BUNIT'])
        except Exception:
            data_unit = u.count

        # FITS WCS is invalid, so ignore it.
        if wcs.spectral.naxis == 0:
            kw = {}
        else:
            kw = {'wcs': wcs}

        return Spectrum1D(flux=data * data_unit, meta=metadata, **kw)

    # Coerce into list-like object
    if (not isinstance(data_obj, (list, tuple, SpectrumCollection)) or
            isinstance(data_obj, fits.HDUList)):
        data_obj = [data_obj]

    # See if this is a multi s2d file
    if app.config == "mosviz" and len(data_obj) == 1 and _check_is_file(data_obj[0]):
        if identify_jwst_s2d_multi_fits("test", data_obj[0]):
            data_obj = SpectrumList.read(data_obj[0])

    # If we're given a string, repeat it for each object
    if isinstance(data_labels, str):
        if len(data_obj) > 1:
            data_labels = [f"{data_labels} {i}" for i in range(len(data_obj))]
        else:
            data_labels = [data_labels]
    elif data_labels is None:
        if len(data_obj) > 1:
            data_labels = [f"2D Spectrum {i}" for i in range(len(data_obj))]
        else:
            data_labels = ['2D Spectrum']

    with app.data_collection.delay_link_manager_update():
        for index, data in enumerate(data_obj):
            # If we got a filepath, first try and parse using the Spectrum1D and
            # SpectrumList parsers, and then fall back to parsing it as a generic
            # FITS file.

            # try parsing file_obj as a URI/URL:
            data = download_uri_to_path(data, cache=cache, local_path=local_path, timeout=timeout)

            if _check_is_file(data):
                try:
                    if ext != 1 or transpose:
                        with fits.open(data) as hdulist:
                            data = _parse_as_spectrum1d(hdulist, ext, transpose)
                    else:
                        data = Spectrum1D.read(data)
                except IORegistryError:
                    with fits.open(data) as hdulist:
                        data = _parse_as_spectrum1d(hdulist, ext, transpose)
            elif isinstance(data, fits.HDUList):
                data = _parse_as_spectrum1d(data, ext, transpose)

            # Make metadata layout conform with other viz.
            data.meta = standardize_metadata(data.meta)

            # Set the instrument
            # TODO: this should not be set to nirspec for all datasets
            data.meta['INSTRUME'] = 'nirspec'

            data.meta['mosviz_row'] = index
            # Get the corresponding label for this data product
            label = data_labels[index]
            app.data_collection[label] = data

        if add_to_table:
            _add_to_table(
                app, data_labels, '2D Spectra',
                table_viewer_reference_name=table_viewer_reference_name
            )

    if show_in_viewer:
        if len(data_labels) > 1:
            raise ValueError("More than one data label provided, unclear " +
                             "which to show in viewer")
        app.add_data_to_viewer(spectrum_2d_viewer_reference_name, data_labels[0])

    return len(data_obj)


def _load_fits_image_from_filename(filename, app):
    with fits.open(filename) as hdulist:
        # We do not use the generated labels
        data_list = [d for d, _ in get_image_data_iterator(app, hdulist, "Image", ext=None)]
    return data_list


@data_parser_registry("mosviz-image-parser")
def mos_image_parser(app, data_obj, data_labels=None, share_image=0,
                     image_viewer_reference_name="image-viewer"):
    """
    Attempts to parse an image-like object or list of images.

    Parameters
    ----------
    app : `~jdaviz.app.Application`
        The application-level object used to reference the viewers.
    data_obj : str or list or image-like
        File path, list, or image-like object to be read as a new row in
        the mosviz table.
    data_labels : str, optional
        The label applied to the glue data component.
    share_image : int, optional
        If 0, images are treated as applying to individual spectra. If non-zero,
        a single image will be shared by multiple spectra so that clicking a
        different row in the table does not reload the displayed image.
        Currently, if non-zero, the provided number must match the number of
        spectra.
    """

    if data_obj is None:
        return

    # The label does not matter here. We overwrite later.
    if isinstance(data_obj, str):
        data_obj = _load_fits_image_from_filename(data_obj, app)
    elif isinstance(data_obj, (list, tuple)) and share_image == 0:
        temp_data = []
        for cur_data_obj in data_obj:
            if isinstance(cur_data_obj, str):
                temp_data += _load_fits_image_from_filename(cur_data_obj, app)
            else:
                data_iter = get_image_data_iterator(app, cur_data_obj, "Image", ext=None)
                temp_data += [d[0] for d in data_iter]
        data_obj = temp_data
    else:
        data_iter = get_image_data_iterator(app, data_obj, "Image", ext=None)
        data_obj = [d[0] for d in data_iter]

    n_data = len(data_obj)
    n_data_range = range(n_data)

    # TODO: Maybe should raise exception?
    if share_image and n_data > 1:  # Just use the first one
        data_obj = [data_obj[0]]

    if data_labels is None:
        if share_image:
            data_labels = ["Shared Image"]
        else:
            data_labels = [f"Image {i}" for i in n_data_range]

    elif isinstance(data_labels, str):
        if share_image:
            data_labels = [data_labels]
        else:
            data_labels = [f"{data_labels} {i}" for i in n_data_range]

    with app.data_collection.delay_link_manager_update():

        for i in n_data_range:
            data_obj[i].label = data_labels[i]
            data_obj[i].meta['mosviz_row'] = i
            app.add_data(data_obj[i], data_labels[i], notify_done=False)

        if share_image:
            # Associate this image with multiple spectra
            data_labels *= share_image
            # Show it on viewer
            app.add_data_to_viewer(image_viewer_reference_name, data_labels[0])

        _add_to_table(app, data_labels, 'Images')


@data_parser_registry("mosviz-metadata-parser")
def mos_meta_parser(app, data_obj=None, ids=None):
    """
    Extracts specific metadata from each data entry's .meta

    Parameters
    ----------
    app : `~jdaviz.app.Application`
        The application-level object used to reference the viewers.
    data_obj : str or list or HDUList
        Optional fallback arg for Identifier column
        File path, list, or an HDUList to extract metadata from if
        metadata isn't found in Data.meta, search for metadata
    ids : list of str
        Optional fallback arg for Identifier column
        A list with identification strings to be used to label mosviz
        table rows. Typically, a list with file names. Used in combination
        with data_obj for fallback
    """

    with app.data_collection.delay_link_manager_update():

        current_columns = [comp.label for comp in app.data_collection['MOS Table'].main_components]
        has_1d = "1D Spectra" in current_columns

        # source name can be taken from 1d spectra
        if has_1d:
            names = _get_source_identifiers(app, "1D Spectra", data_obj, ids)
            _add_to_table(app, names, "Identifier")

        # metadata taken from 2d spectra
        if "2D Spectra" in current_columns:
            filters = query_metadata_by_component(app, "FILTER", "2D Spectra", FALLBACK_NAME)
            gratings = query_metadata_by_component(app, "GRATING", "2D Spectra", FALLBACK_NAME)

            filters_gratings = [(f+'/'+g) for f, g in zip(filters, gratings)]

            _add_to_table(app, filters_gratings, "Filter/Grating")

        # Grab target sky coordinates from 1D spectrum, if possible.
        # This has to happen after Filter/Grating because columns are insertion-ordered.
        if has_1d:
            ra = query_metadata_by_component(app, "SRCRA", "1D Spectra", False)
            dec = query_metadata_by_component(app, "SRCDEC", "1D Spectra", False)
            if all(ra) and all(dec):
                _add_to_table(app, ra, "R.A.")
                _add_to_table(app, dec, "Dec.")

        # Refresh current_columns for Images check
        current_columns = [comp.label for comp in app.data_collection['MOS Table'].main_components]

        # If not in 1D spectrum, source coordinates are taken from image headers, if present.
        if "R.A." not in current_columns and "Images" in current_columns:
            ra = query_metadata_by_component(app, "OBJ_RA", "Images", FALLBACK_NAME)
            dec = query_metadata_by_component(app, "OBJ_DEC", "Images", FALLBACK_NAME)
            _add_to_table(app, ra, "R.A.")
            _add_to_table(app, dec, "Dec.")


def query_metadata_by_component(app, keys, data_type, fallback=FALLBACK_NAME):
    """
    Searches and returns metadata values for a specific data component (type)

    If multiple keys are specified, the first found will be returned based on the order
    keys are provided

    Parameters
    ----------
    app : `~jdaviz.app.Application`
        The application-level object used to reference the viewers.
    keys : str or list
        Metadata keywords to search for
    data_type : str
        The type of data to search for the Source ID keywords in. Can be one of:
        "1D Spectra", "2D Spectra", "Images"
    fallback : str
        The value to return in the event the keyword could not be found
    """

    metadata_list = list()
    # If the user only provided a key that's not already in a container or list, put it in
    # one for the upcoming loop
    if not isinstance(keys, Iterable) or isinstance(keys, str):
        keys = [keys]

    for data in app._jdaviz_helper.get_column(data_type):
        meta = app.data_collection[data].meta

        # Search all given keys to see if they exist. Return the first hit
        key_found = False
        for key in keys:
            if key in meta:
                metadata_list.append(meta.get(key))
                key_found = True
                break
            elif key in meta.get(PRIHDR_KEY, ()):
                metadata_list.append(meta[PRIHDR_KEY].get(key))
                key_found = True
                break

        # If none exist, default to fallback
        if not key_found:
            metadata_list.append(fallback)

    return metadata_list


def _get_source_identifiers(app, data_type, hdus=None, filepaths=None,
                            header_keys=['SOURCEID', 'OBJECT']):
    """
    Helper method to search for the Source Identifier fields.
    Searches the metadata first, then falls back to manually searching hdus (if given any).
    Missing Source IDs should not prevent mosviz from loading

    Parameters
    ----------
    app : `~jdaviz.app.Application`
        The application-level object used to reference the viewers.
    data_type : str
        The type of data to search for the Source ID keywords in. Can be one of:
        "1D Spectra", "2D Spectra", "Images"
    hdus : str or list or HDUList
        Optional fallback arg
        File path, list, or an HDUList to extract metadata from if
        metadata isn't found in Data.meta, search for metadata
    filepaths : list of str
        Optional fallback arg
        A list with identification strings to be used to label mosviz
        table rows. Typically, a list with file names. Used in combination
        with data_obj for fallback
    """
    ids = None
    try:
        ids = query_metadata_by_component(app, keys=header_keys, data_type=data_type)
        if ids is None or set(ids) == set(FALLBACK_NAME):
            raise LookupError
    except Exception:
        # If we fellback for ALL sources, try searching by hdu, if provided any
        if hdus is not None:
            ids = _get_source_identifiers_by_hdu(hdus, filepaths, header_keys)
        # If we weren't given hdus to fallback on, then just return our Fallback value
        else:
            ids = [FALLBACK_NAME] * len(app._jdaviz_helper.get_column(data_type))
    return ids


def _get_source_identifiers_by_hdu(data_obj, filepaths=None,
                                   header_keys=['SOURCEID', 'OBJECT'],
                                   allow_duplicates=False):
    """
    Attempts to extract a list of source identifiers via different header values.
    Searches HDU 0 for each HDUL provided

    Parameters
    ----------
    data_obj : str or list or HDUList
        File path, list, or an HDUList to extract metadata from.
    filepaths : list of str, optional
        A list of filepaths to fallback on if no header values are identified.
    header_keys: str or list of str, optional
        The header value (or values) to search an HDU for to extract the source id.
        List order is assumed to be priority order (i.e. will return first successful
        lookup)
    allow_duplicates : bool
        Flag to allow repeat identifiers. Mostly future proofing in case we eventually
        want to load spectral order 2 as well as 1, and to remind us that we currently
        don't allow that.
    """
    src_names = list()
    # If the user only provided a key that's not already in a container or list, put it in
    # one for the upcoming loop
    if not (isinstance(header_keys, Iterable) and not isinstance(header_keys, (str, dict))):
        header_keys = [header_keys]

    # Prepare HDUs:
    # Coerce into list if needed
    if not isinstance(data_obj, (list, tuple, SpectrumCollection)):
        data_obj = [data_obj]
    hdus = [fits.open(x)[0] if _check_is_file(x) else x for x in data_obj]
    for indx, hdu in enumerate(hdus):
        try:
            # Search all given keys to see if they exist. Return the first hit
            key_found = False
            for key in header_keys:
                if key in hdu.header:
                    key_found = True
                    src_name = hdu.header.get(key)
                    if not allow_duplicates and src_name in src_names:
                        continue
                    else:
                        src_names.append(src_name)
                    break
            # If none exist, default to fallback
            # Fallback 1: filepath if only one is given
            # Fallback 2: filepath at indx, if list of files given
            # Fallback 3: If nothing else, just our fallback value
            if not key_found:
                src_names.append(
                    os.path.basename(filepaths) if type(filepaths) is str
                    else os.path.basename(filepaths[indx]) if type(filepaths) is list
                    else FALLBACK_NAME)
        except Exception:
            # Source ID lookup shouldn't ever prevent target from loading. Downgrade all errors to
            # warnings and use fallback
            warnings.warn("Source name lookup failed for Target " + str(indx) +
                          ". Using fallback ID", RuntimeWarning)
            src_names.append(FALLBACK_NAME)

    return src_names


def _id_files_by_datamodl(label_dict, filepaths, catalog_key=None):
    '''
    Given a dictionary of expected file labels, sort a directory of files into
    their matching categories using their DATAMODL values. The key corresponding
    with the source catalog file must be provided as well.

    Specific to JWST generally (through datamodels) and NIRISS specifically for
    the time, though the eventual plan is to add support for the NIRSpec parser.
    '''
    if catalog_key is None:
        raise ValueError("cat_key must be identified before parsing directory")

    for fp in filepaths:
        if fp.is_dir():
            # Potential names of subdirectories where images are stored
            if fp.name in ("cutouts", "mosviz_cutouts", "images"):
                images = sorted(fp.glob('*.fits*'))
                label_dict['Direct Image'] = images
            else:
                continue

        if fp.suffix in ('.fits', '.fits.gz', '.fit', '.fit.gz'):
            # eligible files will have a DATAMODL value in their primary headers
            header = fits.getheader(fp, ext=0)
            datamodl = header.get('DATAMODL')

            # infer the dispersion direction of 1D and 2D spectra by the last
            # letter of their filter values (e.g., "C" from "GR150C")
            try:
                dispersion = header.get('FILTER')[-1]
                if dispersion not in ('R', 'C'):
                    dispersion = header.get('PUPIL')[-1]
                    if dispersion not in ('R', 'C'):
                        dispersion = None
            except TypeError:
                dispersion = None

            # distinguish image files from counts files -- only the former go
            # through the "Assign World Coordinate System" calibration step
            s_wcs = header.get('S_WCS')

            if datamodl is None:
                continue
            if datamodl == 'MultiSpecModel' and dispersion == 'C':
                label_dict['1D Spectra C'].append(fp)
            elif datamodl == 'MultiSpecModel' and dispersion == 'R':
                label_dict['1D Spectra R'].append(fp)
            elif datamodl == 'MultiSlitModel' and dispersion == 'C':
                label_dict['2D Spectra C'].append(fp)
            elif datamodl == 'MultiSlitModel' and dispersion == 'R':
                label_dict['2D Spectra R'].append(fp)
            elif datamodl == 'ImageModel' and s_wcs is not None:
                label_dict['Direct Image'].append(fp)
            elif datamodl == 'MultiSpecModel' and dispersion is None:
                label_dict['1D Spectra'].append(fp)
            elif datamodl == 'SlitModel' and dispersion is None:
                label_dict['2D Spectra'].append(fp)
            else:
                continue

        elif fp.suffix == '.ecsv':
            if label_dict[catalog_key]:
                raise ValueError("source directory has more "
                                 "than one source catalog")
            else:
                label_dict[catalog_key] = fp

        else:
            continue

    return label_dict


@data_parser_registry("mosviz-niriss-parser")
def mos_niriss_parser(app, data_dir, instrument=None,
                      table_viewer_reference_name='table-viewer'):
    """
    Attempts to parse all data for a NIRISS dataset in a single
    directory, which should include:

    - *_i2d.fits : Direct 2D images
    - *_cat.ecsv : Source catalog
    - *_cal.fits : 2D spectra in vertical (R) and horizontal (C)
        orientations. C spectra are shown first in viewers by default
    - *_x1d.fits : 1D spectra in vertical (R) and horizontal (C)
        orientations. C spectra are shown first in viewers by default

    NOTE: For best performance, it's recommended that your directory
    only contain one source catalog and its associated images/spectra.

    Parameters
    ----------
    app : `~jdaviz.app.Application`
        The application-level object used to reference the viewers.
    data_dir : str
        The path to a directory containing NIRISS data products.
    """
    path = Path(data_dir)
    if not path.is_dir():
        raise ValueError(f"{data_dir} is not a valid directory path")

    # create dict for mapping expected file types to their DATAMODL identifiers
    possible_labels = EXPECTED_FILES[instrument]
    files_by_labels = {k: [] for k in possible_labels}

    # there should only be one source catalog, so that key gets a string
    cat_key = 'Source Catalog'
    files_by_labels[cat_key] = ''

    # use FITS header keywords to sort the directory's files
    files_by_labels = _id_files_by_datamodl(files_by_labels, path.glob('*'),
                                            catalog_key=cat_key)

    # validate that all expected files are present in proper amounts
    _warn_if_not_found(app, files_by_labels)

    # parse relevant information from source catalog if it exists
    cat_id_dict = None
    if files_by_labels[cat_key] != '':
        cat_id_dict = {}
        cat_file = files_by_labels[cat_key]
        try:
            cat_fields = ['label', 'sky_centroid.ra', 'sky_centroid.dec']
            parsed_cat_fields = _fields_from_ecsv(cat_file, cat_fields,
                                                  delimiter=" ")
        except KeyError:
            # Older pipeline builds use different colname to distinguish sources
            cat_fields[0] = 'id'
            parsed_cat_fields = _fields_from_ecsv(cat_file, cat_fields,
                                                  delimiter=" ")
        for row in parsed_cat_fields:
            cat_id_dict[int(row[0])] = (row[1], row[2])

    # set up a dictionary of datasets to add to glue
    add_to_glue = {}

    # Parse images
    image_dict = {}

    if "Direct Image" in files_by_labels:
        print("Loading: Images")
        for image_file in files_by_labels["Direct Image"]:
            # save label for table viewer
            im_split = image_file.stem.split("_")[0]
            if instrument == "niriss":
                pupil = fits.getheader(image_file, ext=0).get('PUPIL')
            elif instrument == "nircam":
                pupil = fits.getheader(image_file, ext=0).get('FILTER')

            image_label = f"Image {im_split} {pupil}"
            image_dict[pupil] = image_label

            # save information from data file
            with fits.open(image_file) as temp:
                data_iter = get_image_data_iterator(app, temp, "Image", ext=None)
                data_obj = [d[0] for d in data_iter]  # We do not use the generated labels
                if len(data_obj) > 1:
                    raise ValueError(f"Found {len(data_obj)} direct images, expected 1.")
                image_data = data_obj[0]

            image_data.label = image_label
            add_to_glue[image_label] = image_data

    # initialize lists of data to be shown in table viewer
    ras = []
    decs = []
    image_add = []
    spec_labels_1d = []
    spec_labels_2d = []

    # Parse 2D spectra
    file_labels_2d = [k for k in files_by_labels.keys() if k.startswith("2D")]

    for flabel in file_labels_2d:
        for fname in files_by_labels[flabel]:
            print(f"Loading: {flabel} sources")
            if flabel in ('2D Spectra R', '2D Spectra C'):
                if instrument == "niriss":
                    filter_name = fits.getheader(fname, ext=0).get('PUPIL')
                elif instrument == "nircam":
                    filter_name = fits.getheader(fname, ext=0).get('FILTER')

                # Orientation denoted by "C", "R", or "C+R" for combined spectra
                orientation = flabel.split()[-1]

            # save HDUs in file that correspond with sources in catalog
            with fits.open(fname, memmap=False) as temp:
                sci_hdus = []
                wav_hdus = {}
                for i in range(len(temp)):
                    if "EXTNAME" in temp[i].header:
                        if temp[i].header["EXTNAME"] == "SCI":
                            if cat_id_dict is not None:
                                if (temp[i].header["SOURCEID"] not in cat_id_dict.keys()):
                                    continue
                            sci_hdus.append(i)
                            wav_hdus[i] = ('WAVELENGTH', temp[i].header['EXTVER'])

                for sci in sci_hdus:
                    if temp[sci].header["SPORDER"] == 1:
                        data = temp[sci].data
                        meta = standardize_metadata(temp[sci].header)
                        meta[PRIHDR_KEY] = standardize_metadata(temp[0].header)

                        # The wavelength is stored in a WAVELENGTH HDU. This is
                        # a 2D array, but in order to be able to use Spectrum1D
                        # we use the average wavelength for all image rows

                        if data.shape[0] > data.shape[1]:
                            # then the input data needs to be transposed, and wavelength
                            # needs to be averaged over axis=1 instead of axis=0
                            data = data.T
                            wav = temp[wav_hdus[sci]].data.mean(axis=1) * u.micron
                        else:
                            wav = temp[wav_hdus[sci]].data.mean(axis=0) * u.micron

                        spec2d = Spectrum1D(data * u.one, spectral_axis=wav, meta=meta)
                        spec2d.meta['INSTRUME'] = instrument.upper()
                        spec2d.meta['mosviz_row'] = len(spec_labels_2d)

                        label = (f"{filter_name} Source "
                                 f"{temp[sci].header['SOURCEID']} spec2d "
                                 f"{orientation}")  # noqa
                        add_to_glue[label] = spec2d

                        # update labels for table viewer
                        if cat_id_dict is not None:
                            ra, dec = cat_id_dict[temp[sci].header["SOURCEID"]]
                            # Store catalog's RA/Dec entry to be available later
                            spec2d.meta['catalog_ra'] = ra
                            spec2d.meta['catalog_dec'] = dec
                            ras.append(ra)
                            decs.append(dec)

                        if filter_name in image_dict:
                            image_add.append(image_dict[filter_name])

                        spec_labels_2d.append(label)

    # Parse 1D spectra
    file_labels_1d = [k for k in files_by_labels.keys() if k.startswith("1D")]

    for flabel in file_labels_1d:
        for fname in files_by_labels[flabel]:
            print(f"Loading: {flabel} sources")

            with fits.open(fname, memmap=False) as temp:
                # Filter out HDUs we care about
                if cat_id_dict is not None:
                    filtered_hdul = fits.HDUList([hdu for hdu in temp if (
                        (hdu.name in ('PRIMARY', 'ASDF')) or
                        (hdu.header.get('SOURCEID', None) in cat_id_dict.keys()))])
                else:
                    filtered_hdul = temp

                # SRCTYPE is required for the specutils JWST x1d reader. The reader will
                # force this to POINT if not set. Under known cases, this field will be set
                # for NIRISS programs; if it's not, something's gone wrong. Catch this
                # warning and reraise as an error to warn users.
                try:
                    with warnings.catch_warnings():
                        warnings.filterwarnings("error",
                                                category=UserWarning,
                                                message=".*SRCTYPE is missing or UNKNOWN*")
                        specs = SpectrumList.read(filtered_hdul, format="JWST x1d multi")
                except UserWarning as e:
                    raise KeyError(f"The SRCTYPE keyword in the header of file {fname} "
                                   "is not populated (expected values: EXTENDED or POINT)") from e

                if instrument == "niriss":
                    filter_name = fits.getheader(fname, ext=0).get('PUPIL')
                elif instrument == "nircam":
                    filter_name = fits.getheader(fname, ext=0).get('FILTER')

                # Orientation denoted by "C", "R", or "C+R" for combined spectra
                orientation = flabel.split()[-1]

                # update 1D labels and standardize metadata for table viewer
                for sp in specs:
                    if (
                        sp.meta['header']['SPORDER'] == 1
                        and sp.meta['header']['EXTNAME'] == 'EXTRACT1D'
                    ):
                        sp.meta = standardize_metadata(sp.meta)
                        sp.meta['mosviz_row'] = len(spec_labels_1d)
                        label = (f"{filter_name} Source "
                                 f"{sp.meta['SOURCEID']} spec1d "
                                 f"{orientation}")
                        spec_labels_1d.append(label)
                        add_to_glue[label] = sp

    # Add the datasets to glue - we do this in one step so that we can easily
    # optimize by avoiding recomputing the full link graph at every add
    with app.data_collection.delay_link_manager_update():
        for label, data in add_to_glue.items():
            app.add_data(data, label, notify_done=False)
        _add_to_table(app, spec_labels_1d, "1D Spectra")
        _add_to_table(app, spec_labels_2d, "2D Spectra")
        _add_to_table(app, image_add, "Images")

        # We then populate the table inside this context manager as
        # _add_to_table does operations that also trigger link manager updates.
        meta_names = _get_source_identifiers(app, "2D Spectra")
        _add_to_table(app, meta_names, "Identifier")
        meta_filters = query_metadata_by_component(app, 'FILTER', "2D Spectra")
        _add_to_table(app, meta_filters, "Filter/Grating")
        meta_ra = query_metadata_by_component(app, 'catalog_ra', "2D Spectra")
        _add_to_table(app, meta_ra, "R.A.")
        meta_dec = query_metadata_by_component(app, 'catalog_dec', "2D Spectra")
        _add_to_table(app, meta_dec, "Dec.")

    app.get_viewer(table_viewer_reference_name)._shared_image = True
